# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for XSD profile generation.

You may want to set the TEST_WRITE_SEDA_FILES environment variable to activate
writing of generated content back to the file-system.
"""

from doctest import Example
from itertools import chain, repeat
import os
from os.path import basename, join
import unittest

from lxml import etree
from lxml.doctestcompare import LXMLOutputChecker

from logilab.common.decorators import cached

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_seda.xsd2yams import XSDMMapping
from cubicweb_seda.entities import profile_generation as pg

from cubicweb_seda import testutils


class XmlTestMixin:
    """Mixin class providing additional assertion methods for checking XML data."""

    NAMESPACES = {
        "xs": "http://www.w3.org/2001/XMLSchema",
        "seda": "fr:gouv:culture:archivesdefrance:seda:v2.0",
        "rng": "http://relaxng.org/ns/structure/1.0",
        "a": "http://relaxng.org/ns/compatibility/annotations/1.0",
    }
    # TBD in concret implementation
    schema_class = None
    schema_file = None
    adapter_id = None

    @classmethod
    @cached
    def schema(cls, xsd_filename):
        with open(xsd_filename) as stream:
            return cls.schema_class(etree.parse(stream))

    def xpath(self, element, expression):
        return element.xpath(expression, namespaces=self.NAMESPACES)

    def get_element(self, profile, name):
        """Return etree element definition for 'name' (there should be only one)"""
        elements = self.get_elements(profile, name)
        self.assertEqual(len(elements), 1)
        return elements[0]

    def get_elements(self, profile, name):
        """Return etree element definitions for 'name'"""
        raise NotImplementedError()

    def get_attribute(self, profile, name):
        """Return etree attribute definition for 'name' (there should be only one)"""
        attributes = self.get_attributes(profile, name)
        self.assertEqual(len(attributes), 1)
        return attributes[0]

    def get_attributes(self, profile, name):
        """Return etree attribute definitions for 'name'"""
        raise NotImplementedError()

    def profile_etree(self, transfer_entity, adapter_id=None):
        """Return etree representation of profile's XSD for the given transfer entity."""
        adapter = transfer_entity.cw_adapt_to(adapter_id or self.adapter_id)
        root = adapter.dump_etree()
        self.assertXmlValid(root)
        return root

    def assertXmlValid(self, root):
        """Validate an XML etree according to an XSD file (.xsd)."""
        schema = self.schema(self.datapath(self.schema_file))
        schema.assert_(root)

    def assertXmlEqual(self, actual, expected):
        """Check that both XML strings represent the same XML tree."""
        checker = LXMLOutputChecker()
        if not checker.check_output(expected, actual, 0):
            message = checker.output_difference(Example("", expected), actual, 0)
            self.fail(message)

    def check_xsd_profile(self, root, sample_file, **substitutions):
        """Check that the SEDA profile can be used to validate a sample XML document."""
        if os.environ.get("TEST_WRITE_SEDA_FILES"):
            fname = join("/tmp", basename(sample_file))
            with open(fname, "w") as stream:
                stream.write(etree.tostring(root, encoding=str, pretty_print=True))
            print(f"Generated profile saved to {fname}")
        profile = self.schema_class(root)
        with open(sample_file) as f:
            sample_xml_string = f.read().format(**substitutions)
        profile.assert_(etree.fromstring(sample_xml_string.encode("utf-8")))


class XMLSchemaTestMixin(XmlTestMixin):
    """Mixin extending XmlTestMixin with implementation of some assert methods for XMLSchema."""

    schema_class = etree.XMLSchema
    schema_file = "XMLSchema.xsd"
    adapter_id = "SEDA-2.1.xsd"

    def get_elements(self, profile, name):
        return self.xpath(profile, f'//xs:element[@name="{name}"]')

    def get_attributes(self, profile, name):
        return self.xpath(profile, f'//xs:attribute[@name="{name}"]')

    def assertElementDefinition(self, element, expected):
        edef = dict(element.attrib)
        types = self.xpath(element, "xs:complexType/xs:simpleContent/xs:extension")
        assert len(types) <= 1
        if types:
            edef["type"] = types[0].attrib["base"]
        self.assertEqual(edef, expected)

    def assertAttributeDefinition(self, element, expected):
        edef = dict(element.attrib)
        edef.setdefault("use", "required")
        self.assertEqual(edef, expected)

    def assertXSDAttributes(self, element, expected_attributes):
        # attributes for regular elements
        attrs = [x.attrib for x in self.xpath(element, "xs:complexType/xs:attribute")]
        # attributes for simple elements
        attrs += [
            x.attrib
            for x in self.xpath(
                element, "xs:complexType/xs:simpleContent/xs:extension/xs:attribute"
            )
        ]
        self.assertEqual(sorted(attrs), expected_attributes)


class RelaxNGTestMixin(XmlTestMixin):
    """Mixin extending XmlTestMixin with implementation of some assert methods for RelaxNG."""

    schema_class = etree.RelaxNG
    schema_file = "relaxng.rng"
    adapter_id = "SEDA-2.1.rng"

    def get_elements(self, profile, name):
        return self.xpath(profile, f'//rng:element[@name="{name}"]')

    def get_attributes(self, profile, name):
        return self.xpath(profile, f'//rng:attribute[@name="{name}"]')

    def assertElementDefinition(self, element, expected):
        edef = dict(element.attrib)  # {name: element name}
        if element.getparent().tag == "{http://relaxng.org/ns/structure/1.0}optional":
            edef["minOccurs"] = "0"
        elif (
            element.getparent().tag == "{http://relaxng.org/ns/structure/1.0}oneOrMore"
        ):
            edef["maxOccurs"] = "unbounded"
        elif (
            element.getparent().tag == "{http://relaxng.org/ns/structure/1.0}zeroOrMore"
        ):
            edef["minOccurs"] = "0"
            edef["maxOccurs"] = "unbounded"
        refs = self.xpath(element, "rng:ref")
        if refs:
            assert len(refs) == 1
            edef["type"] = refs[0].attrib["name"]
        self._element_fixed_value(edef, element)
        self.assertEqual(edef, expected)

    def assertAttributeDefinition(self, element, expected):
        edef = dict(element.attrib)  # {name: element name}
        if element.getparent().tag == "{http://relaxng.org/ns/structure/1.0}optional":
            edef["use"] = "optional"
        else:
            edef["use"] = "required"
        self._element_fixed_value(edef, element)
        self.assertEqual(edef, expected)

    def assertXSDAttributes(self, element, expected_attributes):
        adefs = []
        optattrs = self.xpath(element, "rng:optional/rng:attribute")
        attrs = self.xpath(element, "rng:attribute")
        for use, adef_element in chain(
            zip(repeat("optional"), optattrs), zip(repeat("required"), attrs)
        ):
            adef = dict(adef_element.attrib)
            adef["use"] = use
            data_elements = self.xpath(adef_element, "rng:data")
            if data_elements:
                assert len(data_elements) == 1
                adef["type"] = "xsd:" + data_elements[0].attrib["type"]
            self._element_fixed_value(adef, adef_element)
            adefs.append(adef)
        return sorted(adefs, key=lambda d: list(d.items()))

    def _element_fixed_value(self, edef, element):
        values = self.xpath(element, "rng:value")
        if values:
            assert len(values) == 1
            value = values[0]
            edef["fixed"] = value.text
            if value.attrib.get("type"):
                edef["type"] = "xsd:" + value.attrib["type"]
        else:
            datatypes = self.xpath(element, "rng:data")
            if datatypes:
                assert len(datatypes) == 1
                datatype = datatypes[0]
                edef["type"] = "xsd:" + datatype.attrib["type"]


class PathTargetValuesTC(CubicWebTC):
    def test_keyword_path(self):
        element_defs = iter(XSDMMapping("Keyword"))
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            kw = create("SEDAKeyword", user_cardinality="0..n")
            kt = create("SEDAKeywordType", seda_keyword_type_from=kw)

            edef = next(element_defs)
            # self.assertEqual(
            #  readable_edef(edef),
            #  ('Keyword', 'SEDAKeyword', [
            #      ('id', [('seda_id', 'object', 'SEDAid'),
            #              ('id', 'subject', 'String')]),
            #      ('KeywordContent', []),
            #      ('KeywordReference',
            #       [('seda_keyword_reference_from', 'object', 'SEDAKeywordReference'),
            #        ('seda_keyword_reference_to', 'subject', 'Concept')]),
            #      ('KeywordType', [('seda_keyword_type_from', 'object', 'SEDAKeywordType'),
            #                       ('seda_keyword_type_to', 'subject', 'Concept')]),
            #  ]))
            path = edef[-1][0][1]
            target_values = pg._path_target_values(kw, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0], None)
            self.assertEqual(target_value[1], None)

            path = edef[-1][2][1]
            target_values = pg._path_target_values(kw, path)
            self.assertEqual(len(target_values), 0)

            path = edef[-1][3][1]
            target_values = pg._path_target_values(kw, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0].eid, kt.eid)
            self.assertEqual(target_value[1], None)

            kt_scheme = testutils.scheme_for_rtype(cnx, "seda_keyword_type_to", "theme")
            kw_type = kt_scheme.reverse_in_scheme[0]
            kt.cw_set(seda_keyword_type_to=kw_type)
            path = edef[-1][3][1]
            target_values = pg._path_target_values(kw, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0].eid, kt.eid)
            self.assertEqual(target_value[1].eid, kw_type.eid)

            edef = next(element_defs)
            # self.assertEqual(
            #     readable_edef(edef),
            #     ('KeywordContent', 'SEDAKeyword', [
            #         ('KeywordContent', [('keyword_content', 'subject', 'String')]),
            #     ]))
            path = edef[-1][0][1]
            target_values = pg._path_target_values(kw, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0], None)
            self.assertEqual(target_value[1], None)

    def test_internal_reference(self):
        element_defs = iter(XSDMMapping("DataObjectReference"))
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            transfer = create("SEDAArchiveTransfer", title="test profile")
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            bdo = cnx.create_entity(
                "SEDABinaryDataObject",
                user_annotation="I am mandatory",
                seda_binary_data_object=transfer,
            )
            ref = create(
                "SEDADataObjectReference",
                seda_data_object_reference=unit_alt_seq,
                seda_data_object_reference_id=bdo,
            )

            edef = next(element_defs)
            # readable_edef(edef)
            # ('DataObjectReference',
            #  'SEDADataObjectReference',
            #  [('id', [('seda_id', 'object', 'SEDAid'), ('id', 'subject', 'String')]),
            #   ('DataObjectReferenceId',
            #    [('seda_data_object_reference_id',
            #      'subject',
            #      ('SEDABinaryDataObject', 'SEDAPhysicalDataObject'))])])

            path = edef[-1][1][1]
            target_values = pg._path_target_values(ref, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0], None)
            self.assertEqual(target_value[1].eid, bdo.eid)

    def test_integrity_cardinality(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            bdo = cnx.create_entity(
                "SEDABinaryDataObject",
                user_annotation="I am mandatory",
                seda_binary_data_object=transfer,
            )
            assert pg.integrity_cardinality(bdo) == "1"
            bdo = cnx.create_entity(
                "SEDABinaryDataObject",
                user_cardinality="0..1",
                user_annotation="opt",
                seda_binary_data_object=transfer,
            )
            assert pg.integrity_cardinality(bdo) == "0..1"

            _, _, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx)
            bdo = testutils.create_data_object(unit_alt_seq)
            assert pg.integrity_cardinality(bdo) == "0..n"

            _, _, unit_alt_seq = testutils.create_archive_unit(
                None, cnx=cnx, user_cardinality="1..n"
            )
            bdo = testutils.create_data_object(unit_alt_seq, user_cardinality="0..1")
            assert pg.integrity_cardinality(bdo) == "0..n"

            _, _, unit_alt_seq = testutils.create_archive_unit(
                None, cnx=cnx, user_cardinality="1..n"
            )
            _, _, unit_alt_seq2 = testutils.create_archive_unit(
                unit_alt_seq, cnx=cnx, user_cardinality="0..n"
            )
            bdo = testutils.create_data_object(unit_alt_seq2, user_cardinality="1")
            assert pg.integrity_cardinality(bdo) == "0..n"

            _, _, unit_alt_seq = testutils.create_archive_unit(
                None, cnx=cnx, user_cardinality="1"
            )
            _, _, unit_alt_seq2 = testutils.create_archive_unit(
                unit_alt_seq, cnx=cnx, user_cardinality="1..n"
            )
            bdo = testutils.create_data_object(unit_alt_seq2, user_cardinality="0..1")
            assert pg.integrity_cardinality(bdo) == "0..n"

            _, _, unit_alt_seq = testutils.create_archive_unit(
                None, cnx=cnx, user_cardinality="0..1"
            )
            _, _, unit_alt_seq2 = testutils.create_archive_unit(
                unit_alt_seq, cnx=cnx, user_cardinality="1"
            )
            bdo = testutils.create_data_object(unit_alt_seq2, user_cardinality="1..n")
            assert pg.integrity_cardinality(bdo) == "0..n"


class SEDA2RNGExportTC(RelaxNGTestMixin, CubicWebTC):
    def test_skipped_mandatory_simple(self):
        with self.admin_access.cnx() as cnx:
            profile = self.profile_etree(
                cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            )
            date = self.get_element(profile, "Date")
            self.assertElementDefinition(date, {"name": "Date", "type": "xsd:dateTime"})
            self.assertXSDAttributes(date, [])
            identifier = self.get_element(profile, "MessageIdentifier")
            self.assertElementDefinition(
                identifier, {"name": "MessageIdentifier", "type": "xsd:token"}
            )
            self.assertXSDAttributes(
                identifier,
                [
                    {"name": "schemeAgencyID", "use": "optional", "type": "xsd:token"},
                    {
                        "name": "schemeAgencyName",
                        "use": "optional",
                        "type": "xsd:string",
                    },
                    {"name": "schemeDataURI", "use": "optional", "type": "xsd:anyURI"},
                    {"name": "schemeID", "use": "optional", "type": "xsd:token"},
                    {"name": "schemeName", "use": "optional", "type": "xsd:string"},
                    {"name": "schemeURI", "use": "optional", "type": "xsd:anyURI"},
                    {"name": "schemeVersionID", "use": "optional", "type": "xsd:token"},
                ],
            )

    def test_skipped_mandatory_complex(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            testutils.create_data_object(transfer, filename="fixed.txt")
            profile = self.profile_etree(transfer)
            fname = self.get_element(profile, "Filename")
            self.assertElementDefinition(
                fname, {"name": "Filename", "fixed": "fixed.txt", "type": "xsd:string"}
            )
            self.assertXSDAttributes(fname, [])
            attachment = self.get_element(profile, "Attachment")
            self.assertElementDefinition(
                attachment, {"name": "Attachment", "type": "xsd:base64Binary"}
            )
            self.assertXSDAttributes(
                attachment,
                [
                    {"name": "filename", "use": "optional", "type": "xsd:string"},
                    {"name": "uri", "use": "optional", "type": "xsd:anyURI"},
                ],
            )

    def test_fileinfo_card(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            bdo = cnx.create_entity(
                "SEDABinaryDataObject",
                user_annotation="I am mandatory",
                seda_binary_data_object=transfer,
            )
            appname = cnx.create_entity(
                "SEDACreatingApplicationName", seda_creating_application_name=bdo
            )

            profile = self.profile_etree(transfer)
            fileinfo = self.get_element(profile, "FileInfo")
            self.assertElementDefinition(fileinfo, {"name": "FileInfo"})

            appname.cw_set(user_cardinality="1")
            profile = self.profile_etree(transfer)
            fileinfo = self.get_element(profile, "FileInfo")
            self.assertElementDefinition(fileinfo, {"name": "FileInfo"})

            appname.cw_set(user_cardinality="0..1")
            bdo.cw_set(filename="fixed.txt")
            profile = self.profile_etree(transfer)
            fileinfo = self.get_element(profile, "FileInfo")
            self.assertElementDefinition(fileinfo, {"name": "FileInfo"})

    def test_data_object_package_card(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            bdo = cnx.create_entity(
                "SEDABinaryDataObject",
                user_annotation="I am mandatory",
                seda_binary_data_object=transfer,
            )

            profile = self.profile_etree(transfer)
            fileinfo = self.get_element(profile, "DataObjectPackage")
            self.assertElementDefinition(fileinfo, {"name": "DataObjectPackage"})

            bdo.cw_set(user_cardinality="1")
            profile = self.profile_etree(transfer)
            dop = self.get_element(profile, "DataObjectPackage")
            self.assertElementDefinition(dop, {"name": "DataObjectPackage"})

    def test_object_package_group(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            bdo = cnx.create_entity(
                "SEDABinaryDataObject",
                user_annotation="I am number one",
                seda_binary_data_object=transfer,
            )
            cnx.create_entity(
                "SEDABinaryDataObject",
                user_annotation="I am number two",
                seda_binary_data_object=transfer,
            )
            cnx.create_entity(
                "SEDAPhysicalDataObject",
                user_annotation="I am number three",
                seda_physical_data_object=transfer,
            )

            profile = self.profile_etree(transfer)
            dop = self.get_element(profile, "DataObjectPackage")
            self.assertEqual(len(self.xpath(dop, "./rng:zeroOrMore/rng:choice/*")), 3)

            # setting some cardinality to 1 will remove rng:optional parent of the DataObjectPackage
            # and BinaryDataObject nodes
            bdo.cw_set(user_cardinality="1")
            profile = self.profile_etree(transfer)
            dop = self.get_element(profile, "DataObjectPackage")
            self.assertEqual(len(self.xpath(dop, "./rng:zeroOrMore/rng:choice/*")), 3)

    def test_object_data_object_with_group(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity(
                "SEDAArchiveTransfer", title="test profile", wrap_dataobjects=True
            )
            cnx.create_entity(
                "SEDABinaryDataObject",
                user_cardinality="1",
                user_annotation="I am number one",
                seda_binary_data_object=transfer,
            )
            cnx.create_entity(
                "SEDAPhysicalDataObject",
                user_cardinality="1..n",
                user_annotation="I am number two",
                seda_physical_data_object=transfer,
            )
            cnx.create_entity(
                "SEDABinaryDataObject",
                user_cardinality="0..1",
                user_annotation="I am number one",
                seda_binary_data_object=transfer,
            )
            cnx.create_entity(
                "SEDAPhysicalDataObject",
                user_cardinality="0..n",
                user_annotation="I am number two",
                seda_physical_data_object=transfer,
            )
            profile = self.profile_etree(transfer)
            dop = self.get_element(profile, "DataObjectPackage")
            self.assertEqual(
                len(
                    self.xpath(
                        dop,
                        './rng:zeroOrMore/rng:choice/rng:element[@name="DataObjectGroup"]/'
                        'rng:element[@name="BinaryDataObject"]',
                    )
                ),
                1,
            )
            self.assertEqual(
                len(
                    self.xpath(
                        dop,
                        './rng:zeroOrMore/rng:choice/rng:oneOrMore/rng:element[@name="DataObjectGroup"]'  # noqa
                        '/rng:element[@name="PhysicalDataObject"]',
                    )
                ),
                1,
            )
            self.assertEqual(
                len(
                    self.xpath(
                        dop,
                        './rng:zeroOrMore/rng:choice/rng:zeroOrMore/rng:element[@name="DataObjectGroup"]/'  # noqa
                        'rng:element[@name="PhysicalDataObject"]',
                    )
                ),
                1,
            )
            self.assertEqual(
                len(
                    self.xpath(
                        dop,
                        './rng:zeroOrMore/rng:choice/rng:optional/rng:element[@name="DataObjectGroup"]'  # noqa
                        '/rng:element[@name="BinaryDataObject"]',
                    )
                ),
                1,
            )

    def test_object_data_object_without_group(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity(
                "SEDAArchiveTransfer", title="test profile", wrap_dataobjects=False
            )
            cnx.create_entity(
                "SEDABinaryDataObject",
                user_cardinality="1",
                user_annotation="I am number one",
                seda_binary_data_object=transfer,
            )
            cnx.create_entity(
                "SEDAPhysicalDataObject",
                user_cardinality="1..n",
                user_annotation="I am number two",
                seda_physical_data_object=transfer,
            )
            cnx.create_entity(
                "SEDABinaryDataObject",
                user_cardinality="0..1",
                user_annotation="I am number one",
                seda_binary_data_object=transfer,
            )
            cnx.create_entity(
                "SEDAPhysicalDataObject",
                user_cardinality="0..n",
                user_annotation="I am number two",
                seda_physical_data_object=transfer,
            )

            profile = self.profile_etree(transfer)
            dop = self.get_element(profile, "DataObjectPackage")
            self.assertEqual(
                len(
                    self.xpath(
                        dop,
                        './rng:zeroOrMore/rng:choice/rng:element[@name="BinaryDataObject"]',
                    )
                ),
                1,
            )
            self.assertEqual(
                len(
                    self.xpath(
                        dop,
                        "./rng:zeroOrMore/rng:choice/rng:oneOrMore/"
                        '/rng:element[@name="PhysicalDataObject"]',
                    )
                ),
                1,
            )
            self.assertEqual(
                len(
                    self.xpath(
                        dop,
                        "./rng:zeroOrMore/rng:choice/rng:zeroOrMore/"
                        'rng:element[@name="PhysicalDataObject"]',
                    )
                ),
                1,
            )
            self.assertEqual(
                len(
                    self.xpath(
                        dop,
                        "./rng:zeroOrMore/rng:choice/rng:optional/"
                        '/rng:element[@name="BinaryDataObject"]',
                    )
                ),
                1,
            )

    def test_transfer_annotation(self):
        with self.admin_access.cnx() as cnx:
            profile = self.profile_etree(
                cnx.create_entity(
                    "SEDAArchiveTransfer",
                    title="test profile",
                    user_annotation="some description",
                )
            )
            docs = self.xpath(profile, "///xs:documentation")
            self.assertEqual(len(docs), 1)
            self.assertEqual(docs[0].text, "some description")

    def test_transfer_signature(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            cnx.create_entity("SEDASignature", seda_signature=transfer)
            profile = self.profile_etree(transfer)
            signature = self.get_element(profile, "Signature")
            self.assertElementDefinition(
                signature, {"name": "Signature", "type": "OpenType"}
            )
            self.assertOpenTypeIsDefined(profile)

    def test_keyword(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            scheme = testutils.scheme_for_rtype(cnx, "seda_keyword_type_to", "theme")
            kw_type = scheme.reverse_in_scheme[0]

            transfer = create("SEDAArchiveTransfer", title="test profile")
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer, user_cardinality="0..n"
            )

            kw = create(
                "SEDAKeyword", seda_keyword=unit_alt_seq, keyword_content="kwick"
            )
            kwr_e = create("SEDAKeywordReference", seda_keyword_reference_from=kw)
            create(
                "SEDAKeywordType",
                seda_keyword_type_from=kw,
                seda_keyword_type_to=kw_type,
            )

            profile = self.profile_etree(transfer)

            keywords = self.get_elements(profile, "Keyword")
            self.assertElementDefinition(keywords[0], {"name": "Keyword"})

            kwc = self.get_elements(profile, "KeywordContent")
            self.assertElementDefinition(
                kwc[0],
                {"name": "KeywordContent", "type": "xsd:string", "fixed": "kwick"},
            )

            kwt = self.get_element(profile, "KeywordType")
            self.assertElementDefinition(
                kwt, {"name": "KeywordType", "type": "xsd:token", "fixed": "theme"}
            )
            self.assertXSDAttributes(
                kwt,
                [
                    {
                        "name": "listVersionID",
                        "fixed": "seda_keyword_type_to/None vocabulary",
                    }
                ],
            )
            kwr = self.get_element(profile, "KeywordReference")
            self.assertElementDefinition(
                kwr, {"name": "KeywordReference", "type": "xsd:token"}
            )
            self.assertXSDAttributes(
                kwr,
                [
                    {"name": "schemeAgencyID", "type": "xsd:token", "use": "optional"},
                    {
                        "name": "schemeAgencyName",
                        "type": "xsd:string",
                        "use": "optional",
                    },
                    {"name": "schemeDataURI", "type": "xsd:anyURI", "use": "optional"},
                    {"name": "schemeID", "type": "xsd:token", "use": "optional"},
                    {"name": "schemeName", "type": "xsd:string", "use": "optional"},
                    {"name": "schemeURI", "type": "xsd:anyURI", "use": "optional"},
                    {"name": "schemeVersionID", "type": "xsd:token", "use": "optional"},
                ],
            )

            kwr_e.cw_set(seda_keyword_reference_to_scheme=scheme)
            profile = self.profile_etree(transfer)
            kwr = self.get_element(profile, "KeywordReference")
            self.assertElementDefinition(
                kwr, {"name": "KeywordReference", "type": "xsd:token"}
            )
            self.assertXSDAttributes(
                kwr, [{"name": "schemeURI", "fixed": scheme.cwuri}]
            )

            kwr_e.cw_set(seda_keyword_reference_to=kw_type)
            profile = self.profile_etree(transfer)
            kwr = self.get_element(profile, "KeywordReference")
            self.assertElementDefinition(
                kwr,
                {
                    "name": "KeywordReference",
                    "type": "xsd:token",
                    "fixed": kw_type.cwuri,
                },
            )
            self.assertXSDAttributes(
                kwr, [{"name": "schemeURI", "fixed": scheme.cwuri}]
            )

    def test_code_list(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            scheme = cnx.create_entity("ConceptScheme", title="Keyword Types")
            cnx.create_entity(
                "SEDAMimeTypeCodeListVersion",
                seda_mime_type_code_list_version_from=transfer,
                seda_mime_type_code_list_version_to=scheme,
            )

            profile = self.profile_etree(transfer)
            elements = self.get_elements(profile, "MimeTypeCodeListVersion")
            self.assertEqual(len(elements), 0)

    def test_multiple_concepts(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            scheme = testutils.scheme_for_type(
                cnx, None, None, "application/msword", "application/pdf"
            )
            cnx.create_entity(
                "SEDAMimeTypeCodeListVersion",
                seda_mime_type_code_list_version_from=transfer,
                seda_mime_type_code_list_version_to=scheme,
            )
            bdo = testutils.create_data_object(transfer)
            bdo.mime_type.cw_set(
                user_cardinality="1", seda_mime_type_to=scheme.reverse_in_scheme
            )

            profile = self.profile_etree(transfer)
            mt = self.get_element(profile, "MimeType")
            self.assertEqual(
                "\n".join(
                    etree.tostring(mt, encoding=str, pretty_print=True).splitlines()[
                        1:-1
                    ]
                ),
                """\
  <rng:choice>
    <rng:value type="token">application/msword</rng:value>
    <rng:value type="token">application/pdf</rng:value>
  </rng:choice>""",
            )

    def test_seda2_concept(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            scheme = create("ConceptScheme", title="Digest algorithm")
            some_concept = scheme.add_concept(label="md5 algorithm", language_code="en")
            transfer = create("SEDAArchiveTransfer", title="test profile")
            cnx.create_entity(
                "SEDAMessageDigestAlgorithmCodeListVersion",
                seda_message_digest_algorithm_code_list_version_from=transfer,
                seda_message_digest_algorithm_code_list_version_to=scheme,
            )
            create(
                "SEDABinaryDataObject",
                user_cardinality="0..n",
                user_annotation="I am mandatory",
                seda_binary_data_object=transfer,
                seda_algorithm=some_concept,
            )

            profile = self.profile_etree(transfer)
            algo = self.get_attribute(profile, "algorithm")
            self.assertAttributeDefinition(
                algo,
                {
                    "name": "algorithm",
                    "use": "required",
                    "type": "xsd:token",
                    "fixed": "md5 algorithm",
                },
            )

            create(
                "Label",
                label_of=some_concept,
                kind="preferred",
                language_code="seda-2",
                label="md5",
            )

            some_concept.cw_clear_all_caches()
            profile = self.profile_etree(transfer)
            algo = self.get_attribute(profile, "algorithm")
            self.assertAttributeDefinition(
                algo,
                {
                    "name": "algorithm",
                    "use": "required",
                    "type": "xsd:token",
                    "fixed": "md5",
                },
            )

    def assertOpenTypeIsDefined(self, profile):
        open_types = self.xpath(profile, '//rng:define[@name="OpenType"]')
        self.assertEqual(len(open_types), 1)

    def test_data_duplicates(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer, user_cardinality="0..n"
            )
            profile = self.profile_etree(transfer)
            title = self.get_element(profile, "Title")
            self.assertEqual(len(self.xpath(title, "rng:data")), 1)


class SEDAExportFuncTCMixIn:
    """Test that SEDA profile export works correctly."""

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            scheme = create("ConceptScheme", title="Keyword Types")
            # ensure we're able to export concept with unexpected language code
            some_concept = scheme.add_concept(label="md5", language_code="de")
            some_other_concept = scheme.add_concept(
                label="hash/md5", language_code="de"
            )

            transfer = create("SEDAArchiveTransfer", title="test profile")
            create(
                "SEDAMessageDigestAlgorithmCodeListVersion",
                seda_message_digest_algorithm_code_list_version_from=transfer,
                seda_message_digest_algorithm_code_list_version_to=scheme,
            )
            create(
                "SEDAFileFormatCodeListVersion",
                seda_file_format_code_list_version_from=transfer,
            )
            create(
                "SEDAMimeTypeCodeListVersion",
                user_cardinality="0..1",
                seda_mime_type_code_list_version_from=transfer,
                seda_mime_type_code_list_version_to=scheme,
            )
            access_rule = create(
                "SEDAAccessRule", user_cardinality="0..1", seda_access_rule=transfer
            )
            access_rule_seq = create(
                "SEDASeqAccessRuleRule", reverse_seda_seq_access_rule_rule=access_rule
            )
            create(
                "SEDAStartDate",
                user_cardinality="0..1",
                seda_start_date=access_rule_seq,
            )
            # binary data object
            bdo = testutils.create_data_object(
                transfer, user_cardinality="0..n", seda_algorithm=some_concept
            )
            bdo.mime_type.cw_set(seda_mime_type_to=[some_concept, some_other_concept])
            create(
                "SEDAFormatLitteral", user_cardinality="0..1", seda_format_litteral=bdo
            )
            create("SEDAEncoding", user_cardinality="0..1", seda_encoding_from=bdo)
            create("SEDAUri", seda_uri=bdo.seda_alt_binary_data_object_attachment)
            # first level archive unit
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer, user_cardinality="0..n", user_annotation="Composant ISAD(G)"
            )
            # management rules
            appraisal_rule = create(
                "SEDAAppraisalRule", seda_appraisal_rule=unit_alt_seq
            )
            appraisal_rule_seq = create(
                "SEDASeqAppraisalRuleRule",
                reverse_seda_seq_appraisal_rule_rule=appraisal_rule,
            )
            create(
                "SEDAStartDate",
                user_cardinality="0..1",
                seda_start_date=appraisal_rule_seq,
            )
            access_rule = create("SEDAAccessRule", seda_access_rule=unit_alt_seq)
            create("SEDADisseminationRule", seda_dissemination_rule=unit_alt_seq)
            create("SEDAReuseRule", seda_reuse_rule=unit_alt_seq)
            create("SEDANeedAuthorization", seda_need_authorization=unit_alt_seq)
            # content
            kw = create(
                "SEDAKeyword", user_cardinality="0..n", seda_keyword=unit_alt_seq
            )
            create("SEDAKeywordType", seda_keyword_type_from=kw)
            create("SEDAKeywordReference", seda_keyword_reference_from=kw)
            history_item = create(
                "SEDACustodialHistoryItem", seda_custodial_history_item=unit_alt_seq
            )
            create("SEDAwhen", user_cardinality="0..1", seda_when=history_item)
            version_of = create("SEDAIsVersionOf", seda_is_version_of=unit_alt_seq)
            alt2 = create(
                "SEDAAltIsVersionOfArchiveUnitRefId",
                reverse_seda_alt_is_version_of_archive_unit_ref_id=version_of,
            )
            create("SEDADataObjectReference", seda_data_object_reference=alt2)
            create(
                "SEDAOriginatingAgencyArchiveUnitIdentifier",
                seda_originating_agency_archive_unit_identifier=unit_alt_seq,
            )
            create(
                "SEDATransferringAgencyArchiveUnitIdentifier",
                seda_transferring_agency_archive_unit_identifier=unit_alt_seq,
            )
            create("SEDADescription", seda_description=unit_alt_seq)
            create("SEDALanguage", seda_language_from=unit_alt_seq)
            create(
                "SEDADescriptionLanguage", seda_description_language_from=unit_alt_seq
            )
            create("SEDACreatedDate", seda_created_date=unit_alt_seq)
            create("SEDAEndDate", seda_end_date=unit_alt_seq)
            create(
                "SEDADataObjectReference",
                user_cardinality="0..n",
                seda_data_object_reference=unit_alt_seq,
                seda_data_object_reference_id=bdo,
            )

            cnx.commit()
        self.transfer_eid = transfer.eid
        self.bdo_eid = bdo.eid
        self.bdo_xmlid = bdo.cw_adapt_to("IXmlId").id()
        self.au_eid = unit.eid

    def test_profile1(self):
        """Check a minimal SEDA profile validating BV2.1_min.xml."""
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            root = self.profile_etree(transfer)
        self.check_xsd_profile(root, self.datapath("BV2.1_min.xml"))
        # ensure jumped element without content are not there
        self.assertEqual(len(self.get_elements(root, "Gps")), 0)
        # ensure element with skipped value are not there
        self.assertEqual(len(self.get_elements(root, "TransactedDate")), 0)
        self.assertProfileDetails(root)


class SEDARNGExportFuncTC(SEDAExportFuncTCMixIn, RelaxNGTestMixin, CubicWebTC):
    def assertProfileDetails(self, root):
        # ensure profile's temporary id are exported in xml:id attribute
        self.assertEqual(len(self.xpath(root, "//rng:attribute[@xml:id]")), 2)
        for attrdef in self.xpath(root, "//xs:attribute[@xml:id]"):
            self.assertEqual(attrdef[0]["type"], "ID")
        # ensure they are properly referenced using 'default' attribute
        references = self.xpath(
            root, f'//rng:element[@a:defaultValue="{self.bdo_xmlid}"]'
        )
        self.assertEqual(len(references), 1)
        self.assertEqual(references[0].attrib["name"], "DataObjectReferenceId")
        for reference_id in self.xpath(
            root, '//rng:element[@name="DataObjectReferenceId"]'
        ):
            self.assertEqual(reference_id[0].attrib["type"], "NCName")
        # ensure optional id are properly reinjected
        references = self.xpath(
            root,
            '//rng:element[@name="Keyword"]/rng:optional' '/rng:attribute[@name="id"]',
        )
        self.assertEqual(len(references), 1)
        # ensure custodial item content type is properly serialized
        chi = self.xpath(root, '//rng:element[@name="CustodialHistoryItem"]')
        self.assertEqual(len(chi), 1)
        self.assertXSDAttributes(
            chi[0], [{"name": "when", "use": "optional", "type": "datedateTime"}]
        )
        # ensure types union handling
        ddt = self.xpath(root, '//rng:element[@name="CreatedDate"]/rng:choice/rng:data')
        self.assertEqual(len(ddt), 7)
        self.assertEqual(
            {elmt.attrib["type"] for elmt in ddt},
            {
                "date",
                "dateTime",
                "gDay",
                "gYearMonth",
                "gMonthDay",
                "gMonth",
                "gYear",
            },
        )


class OldSEDAExportMixin:
    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            concepts = {}
            for rtype, etype, labels in [
                ("file_category", None, ["fmt/123", "fmt/987"]),
                ("seda_encoding_to", None, ["6"]),
                ("seda_type_to", None, ["CDO"]),
                ("seda_description_level_to", None, ["file"]),
                ("seda_algorithm", "SEDABinaryDataObject", ["md5"]),
                ("seda_rule", "SEDASeqAppraisalRuleRule", ["P10Y"]),
                ("seda_rule", "SEDASeqAccessRuleRule", ["AR038"]),
                ("seda_final_action", "SEDAAppraisalRule", ["detruire"]),
            ]:
                scheme = testutils.scheme_for_type(cnx, rtype, etype, *labels)
                if len(labels) == 1:
                    concepts[labels[0]] = scheme.reverse_in_scheme[0]
                else:
                    for concept in scheme.reverse_in_scheme:
                        concepts[concept.label()] = concept

            # ensure we're able to export concept with unexpected language code
            concepts["md5"].preferred_label[0].cw_set(language_code="de")

            agent = testutils.create_authority_record(
                cnx, "bob", record_id="a_record_id"
            )

            transfer = create(
                "SEDAArchiveTransfer",
                title="my profile title &&",
                simplified_profile=True,
            )

            create(
                "SEDAComment",
                user_cardinality="1",
                comment="my profile description &&",
                seda_comment=transfer,
            )

            create(
                "SEDAAccessRule",  # XXX mandatory for seda 1.0
                user_cardinality="1",
                seda_access_rule=transfer,
                seda_seq_access_rule_rule=create(
                    "SEDASeqAccessRuleRule",
                    reverse_seda_start_date=create("SEDAStartDate"),
                ),
            )
            appraisal_rule_rule = create(
                "SEDASeqAppraisalRuleRule",
                seda_rule=concepts["P10Y"],
                user_annotation="C'est dans 10ans je m'en irai",
                reverse_seda_start_date=create("SEDAStartDate"),
            )
            create(
                "SEDAAppraisalRule",
                user_cardinality="0..1",
                seda_appraisal_rule=transfer,
                seda_final_action=concepts["detruire"],
                seda_seq_appraisal_rule_rule=appraisal_rule_rule,
                user_annotation="detruire le document",
            )

            unit, _, unit_alt_seq = testutils.create_archive_unit(
                transfer, user_cardinality="1..n"
            )

            unit_alt_seq.cw_set(
                reverse_seda_start_date=create(
                    "SEDAStartDate", user_cardinality="0..1"
                ),
                reverse_seda_end_date=create("SEDAEndDate", user_cardinality="0..1"),
                # XXX, value=date(2015, 2, 24)),
                reverse_seda_description=create(
                    "SEDADescription", user_cardinality="0..1"
                ),
            )

            kw = create(
                "SEDAKeyword", user_cardinality="0..n", seda_keyword=unit_alt_seq
            )
            create(
                "SEDAKeywordReference",
                seda_keyword_reference_from=kw,
                seda_keyword_reference_to=concepts["file"],
                seda_keyword_reference_to_scheme=concepts["file"].scheme,
            )
            create(
                "SEDAKeywordType", user_cardinality="0..1", seda_keyword_type_from=kw
            )

            create(
                "SEDAOriginatingAgency",
                user_cardinality="0..1",
                seda_originating_agency_from=unit_alt_seq,
                seda_originating_agency_to=agent,
            )

            create(
                "SEDAType",
                user_cardinality="0..1",
                seda_type_from=unit_alt_seq,
                seda_type_to=concepts["CDO"],
            )

            chi = create(
                "SEDACustodialHistoryItem",
                user_cardinality="0..1",
                seda_custodial_history_item=unit_alt_seq,
                reverse_seda_when=create("SEDAwhen", user_cardinality="0..1"),
            )

            # Add a sub archive unit
            subunit2, _, subunit2_alt_seq = testutils.create_archive_unit(
                unit_alt_seq, user_cardinality="1"
            )

            # Add another sub archive unit
            subunit1, _, subunit_alt_seq = testutils.create_archive_unit(
                unit_alt_seq, user_cardinality="1..n"
            )

            create(
                "SEDAAppraisalRule",
                user_cardinality="0..1",
                seda_appraisal_rule=subunit_alt_seq,
                seda_seq_appraisal_rule_rule=create(
                    "SEDASeqAppraisalRuleRule",
                    reverse_seda_start_date=create("SEDAStartDate"),
                ),
            )

            create(
                "SEDAAccessRule",
                user_cardinality="1",
                user_annotation="restrict",
                seda_access_rule=subunit_alt_seq,
                seda_seq_access_rule_rule=create(
                    "SEDASeqAccessRuleRule",
                    reverse_seda_start_date=create("SEDAStartDate"),
                    seda_rule=concepts["AR038"],
                ),
            )

            # Add minimal document to first level archive
            ref = create(
                "SEDADataObjectReference",
                user_cardinality="0..1",
                seda_data_object_reference=unit_alt_seq,
            )
            bdo = testutils.create_data_object(
                transfer,
                user_cardinality="0..n",
                filename="this_is_the_filename.pdf",
                seda_algorithm=concepts["md5"],
                reverse_seda_data_object_reference_id=ref,
            )

            bdo.format_id.cw_set(
                user_cardinality="1",
                seda_format_id_to=[concepts["fmt/123"], concepts["fmt/987"]],
            )
            create(
                "SEDAEncoding",
                user_cardinality="1",
                seda_encoding_from=bdo,
                seda_encoding_to=concepts["6"],
            )

            cnx.commit()

        self.transfer_eid = transfer.eid
        self.unit_eid = unit.eid
        self.subunit1_eid = subunit1.eid
        self.subunit2_eid = subunit2.eid
        self.bdo_eid = bdo.eid
        self.kw_eid = kw.eid
        self.chi_eid = chi.eid
        self.file_concept_eid = concepts["file"].eid
        self.agent_eid = agent.eid

    def _test_profile(self, adapter_id, expected_file):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            file_concept = cnx.entity_from_eid(self.file_concept_eid)
            agent = cnx.entity_from_eid(self.agent_eid)

            adapter = transfer.cw_adapt_to(adapter_id)
            generated_xsd = adapter.dump(_encoding=str)

            if os.environ.get("TEST_WRITE_SEDA_FILES"):
                orig_content = generated_xsd
                for value, key in [
                    (file_concept.cwuri, "concept-uri"),
                    (file_concept.scheme.cwuri, "scheme-uri"),
                    (str(agent.eid), "agent-id"),
                    (agent.dc_title(), "agent-name"),
                ]:
                    orig_content = orig_content.replace(value, f"%({key})s")
                with open(self.datapath(expected_file + ".new"), "w") as stream:
                    stream.write(orig_content)
                print(f"Regenerated expected file as {expected_file}.new")
            root = etree.fromstring(generated_xsd.encode("utf-8"))
            self.assertXmlValid(root)
            with open(self.datapath(expected_file), "r") as f:
                expected = f.read() % {
                    "unit-eid": str(self.unit_eid),
                    "subunit1-eid": str(self.subunit1_eid),
                    "subunit2-eid": str(self.subunit2_eid),
                    "bdo-eid": str(self.bdo_eid),
                    "kw-eid": str(self.kw_eid),
                    "chi-eid": str(self.chi_eid),
                    "concept-uri": file_concept.cwuri,
                    "scheme-uri": file_concept.scheme.cwuri,
                    "agent-id": str(agent.eid),
                    "agent-name": agent.dc_title(),
                }
            self.assertXmlEqual(expected, generated_xsd)
            return adapter, root


class OldSEDAXSDExportTC(XMLSchemaTestMixin, OldSEDAExportMixin, CubicWebTC):
    def test_seda_1_0(self):
        self._test_profile("SEDA-1.0.xsd", "seda_1_export.xsd")

    def test_seda_0_2(self):
        self._test_profile("SEDA-0.2.xsd", "seda_02_export.xsd")

    def _test_profile(self, adapter_id, expected_file):
        adapter, root = super()._test_profile(adapter_id, expected_file)
        # ensure there is no element with @type but a complex type
        namespaces = adapter.namespaces.copy()
        namespaces.pop(None)
        dates = root.xpath(
            '//xsd:element[@name="Date"]/xsd:complexType/xsd:sequence',
            namespaces=namespaces,
        )
        self.assertEqual(len(dates), 0)


class OldSEDARNGExportTC(RelaxNGTestMixin, OldSEDAExportMixin, CubicWebTC):
    def test_seda_1_0(self):
        self._test_profile("SEDA-1.0.rng", "seda_1_export.rng")

    def test_seda_0_2_rng(self):
        self._test_profile("SEDA-0.2.rng", "seda_02_export.rng")


class OldSEDAExportTC(RelaxNGTestMixin, CubicWebTC):
    def test_seda_0_2_bordereau_ref(self):
        """Check a sample SEDA 0.2 profile validation."""
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            transfer = create(
                "SEDAArchiveTransfer", title="test profile", simplified_profile=True
            )
            create("SEDAComment", seda_comment=transfer)

            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            create("SEDAArchivalAgreement", seda_archival_agreement=transfer)
            create(
                "SEDATransferringAgencyArchiveUnitIdentifier",
                seda_transferring_agency_archive_unit_identifier=unit_alt_seq,
            )
            create("SEDAStartDate", seda_start_date=unit_alt_seq)
            create("SEDAEndDate", seda_end_date=unit_alt_seq)
            appraisal_rule = create(
                "SEDAAppraisalRule",
                user_cardinality="0..1",
                seda_appraisal_rule=unit_alt_seq,
            )
            appraisal_rule_seq = create(
                "SEDASeqAppraisalRuleRule",
                reverse_seda_seq_appraisal_rule_rule=appraisal_rule,
            )
            create("SEDAStartDate", seda_start_date=appraisal_rule_seq)
            access_rule = create(
                "SEDAAccessRule",
                user_cardinality="0..1",
                seda_access_rule=unit_alt_seq,
            )
            access_rule_seq = create(
                "SEDASeqAccessRuleRule", reverse_seda_seq_access_rule_rule=access_rule
            )
            create("SEDAStartDate", seda_start_date=access_rule_seq)

            subunit, subunit_alt, subunit_alt_seq = testutils.create_archive_unit(
                unit_alt_seq
            )
            create(
                "SEDATransferringAgencyArchiveUnitIdentifier",
                seda_transferring_agency_archive_unit_identifier=subunit_alt_seq,
            )
            create("SEDAStartDate", seda_start_date=subunit_alt_seq)
            create("SEDAEndDate", seda_end_date=subunit_alt_seq)
            create("SEDADescription", seda_description=subunit_alt_seq)
            kw = create(
                "SEDAKeyword", user_cardinality="0..n", seda_keyword=subunit_alt_seq
            )
            create("SEDAKeywordReference", seda_keyword_reference_from=kw)

            create("SEDASystemId", seda_system_id=subunit_alt_seq)

            bdo = testutils.create_data_object(transfer, user_cardinality="0..1")
            create(
                "SEDADataObjectReference",
                seda_data_object_reference=subunit_alt_seq,
                seda_data_object_reference_id=bdo,
            )
            create("SEDAEncoding", seda_encoding_from=bdo)
            bdo.mime_type.cw_set(user_cardinality="1")
            create("SEDADateCreatedByApplication", seda_date_created_by_application=bdo)

            cnx.commit()

            root = self.profile_etree(transfer, "SEDA-0.2.rng")
        self.check_xsd_profile(root, self.datapath("seda_02_bordereau_ref.xml"))

    def test_xsd_children_order(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            transfer = create(
                "SEDAArchiveTransfer", title="test profile", simplified_profile=True
            )
            create(
                "SEDAAccessRule",  # XXX mandatory for seda 1.0
                user_cardinality="1",
                seda_access_rule=transfer,
                seda_seq_access_rule_rule=create(
                    "SEDASeqAccessRuleRule",
                    reverse_seda_start_date=create("SEDAStartDate"),
                ),
            )
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer, user_cardinality="0..1"
            )
            testutils.create_archive_unit(transfer)
            testutils.create_archive_unit(unit_alt_seq)
            bdo = testutils.create_data_object(transfer)
            create(
                "SEDADataObjectReference",
                seda_data_object_reference=unit_alt_seq,
                seda_data_object_reference_id=bdo,
            )
            cnx.commit()

            # ensure Document appears before Contains in SEDA 0.2
            adapter = transfer.cw_adapt_to("SEDA-0.2.xsd")
            root = etree.Element("test-root")
            adapter.xsd_children(root, unit)
            self.assertEqual(
                [node.attrib["name"] for node in root], ["Document", "Contains"]
            )

            # ensure Document appears after ArchiveObject in SEDA 1
            adapter = transfer.cw_adapt_to("SEDA-1.0.xsd")
            root = etree.Element("test-root")
            adapter.xsd_children(root, unit)
            self.assertEqual(
                [node.attrib["name"] for node in root], ["ArchiveObject", "Document"]
            )

    def test_duplicated_format_id(self):
        with self.admin_access.cnx() as cnx:
            scheme = testutils.scheme_for_type(
                cnx, "file_category", None, "fmt/123", "fmt/123"
            )
            concepts = scheme.reverse_in_scheme

            transfer = cnx.create_entity(
                "SEDAArchiveTransfer", title="my profile", simplified_profile=True
            )
            bdo = testutils.create_data_object(transfer)
            bdo.format_id.cw_set(seda_format_id_to=concepts)

            adapter = transfer.cw_adapt_to("SEDA-1.0.rng")
            root = etree.Element("test-root")
            adapter.xsd_attachment(root, bdo)

            self.assertEqual(
                [node.text for node in self.xpath(root, "//rng:value")], ["fmt/123"]
            )


class SEDAExportUnitTest(unittest.TestCase):
    def test_concepts_languages(self):
        self.assertEqual(pg.SEDA1XSDExport.concepts_language, "seda-1")
        self.assertEqual(pg.SEDA02XSDExport.concepts_language, "seda-02")

    def test_concept_value(self):
        class concept:
            labels = {}

        concept.labels["fr"] = "Bonjour"
        self.assertEqual(pg._concept_value(concept, "seda-1"), "Bonjour")
        concept.labels["en"] = "Hello"
        self.assertEqual(pg._concept_value(concept, "seda-1"), "Hello")
        concept.labels["seda"] = "hello"
        self.assertEqual(pg._concept_value(concept, "seda-1"), "hello")
        concept.labels["seda-1"] = "hello you"
        self.assertEqual(pg._concept_value(concept, "seda-1"), "hello you")
        concept.labels["seda-2"] = "good-by"
        self.assertEqual(pg._concept_value(concept, "seda-1"), "hello you")


if __name__ == "__main__":
    unittest.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for XSD profile generation"""

from collections import namedtuple

from lxml import etree

from cubicweb.devtools.testlib import CubicWebTC

from test_profile_generation import XmlTestMixin, SEDAExportFuncTCMixIn


AttrDef = namedtuple("AttrDef", ["label", "card", "value"])


class SEDAHTMLExportFuncTC(SEDAExportFuncTCMixIn, XmlTestMixin, CubicWebTC):
    adapter_id = "SEDA-2.1.html"

    def setup_database(self):
        super().setup_database()
        with self.admin_access.cnx() as cnx:
            scheme = cnx.find("ConceptScheme", title="Keyword Types").one()
            some_concept = scheme.reverse_in_scheme[0]
            lang_rtype = cnx.find("CWRType", name="seda_language_to").one()
            some_concept.scheme.cw_set(scheme_relation_type=lang_rtype)
            cnx.find("SEDALanguage").one().cw_set(seda_language_to=some_concept)
            cnx.find("SEDAKeywordReference").one().cw_set(
                seda_keyword_reference_to=some_concept,
                seda_keyword_reference_to_scheme=some_concept.scheme,
            )
            cnx.commit()

        self.concept_eid = some_concept.eid

    def assertXmlValid(self, root):
        pass  # no schema available for html 5

    def get_element(self, profile, name):
        elements = self.xpath(profile, f'//h3[text()="{name}"]')
        self.assertEqual(len(elements), 1)
        return elements[0].getparent()

    def element_values(self, element):
        el_defs = []
        for div in self.xpath(element, "div"):
            el_def = {}
            el_defs.append(el_def)
            for span in self.xpath(div, "span"):
                if len(span):
                    value = etree.tostring(span[0], encoding=str)
                else:
                    value = span.text
                el_def[span.attrib["class"]] = value
        return el_defs

    def test_profile1(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            some_concept = cnx.entity_from_eid(self.concept_eid)
            profile = self.profile_etree(transfer)
            # ensure encoding is declared
            meta = self.xpath(profile, "/html/head/meta")[0]
            self.assertEqual(
                meta.attrib,
                {"content": "text/html; charset=UTF-8", "http-equiv": "content-type"},
            )
            # ensure only one first level title
            self.assertEqual(len(self.xpath(profile, "//title")), 1)
            self.assertEqual(len(self.xpath(profile, "//h1")), 1)
            # ensure annotation are serialized
            self.assertEqual(
                [
                    ann
                    for ann in self.xpath(profile, '//div[@class="description"]/text()')
                    if ann != "data object title"
                ],
                ["Composant ISAD(G)"],
            )
            # ensure all attributes have label, card and value defined
            attr_divs = self.xpath(profile, '//div[@class="attribute"]')
            attr_defs = set()
            for attr_div in attr_divs:
                attr_def = AttrDef(
                    self.xpath(attr_div, 'span[@class="label"]')[0].text,
                    self.xpath(attr_div, 'span[@class="card"]')[0].text,
                    self.xpath(attr_div, 'span[@class="value"]')[0].text,
                )
                attr_defs.add(attr_def)
            self.assertEqual(
                attr_defs,
                {
                    AttrDef(label="algorithm", card="mandatory", value=None),
                    AttrDef(label="filename", card="optional", value=None),
                    AttrDef(
                        label="id",
                        card="mandatory",
                        value=f"id{self.au_eid}",
                    ),
                    AttrDef(
                        label="id",
                        card="mandatory",
                        value=f"id{self.bdo_eid}",
                    ),
                    AttrDef(label="id", card="optional", value=None),
                    AttrDef(label="when", card="optional", value=None),
                    AttrDef(label="uri", card="optional", value=None),
                },
            )
            # ensure KeywordType list handling
            ktype = self.get_element(profile, "KeywordType")
            el_defs = self.element_values(ktype)
            self.assertEqual(
                el_defs,
                [
                    {"label": "listVersionID", "value": "edition 2009"},
                    {"label": "XSD content type", "value": "xsd:token"},
                ],
            )
            # KeywordReference
            kref = self.get_element(profile, "KeywordReference")
            el_defs = self.element_values(kref)
            concept_link = f'<a href="{some_concept.absolute_url()}">md5</a>'
            scheme = some_concept.scheme
            scheme_link = f'<a href="{scheme.absolute_url()}">Keyword Types</a>'
            self.assertEqual(
                el_defs,
                [
                    {"label": "schemeURI", "value": scheme_link},
                    {"label": "XSD content type", "value": "xsd:token"},
                    {"label": "fixed value", "value": concept_link},
                ],
            )
            # Language
            lang = self.get_element(profile, "Language")
            el_defs = self.element_values(lang)
            self.assertEqual(
                el_defs,
                [
                    {"label": "XSD content type", "value": "xsd:language"},
                    {"label": "fixed value", "value": concept_link},
                ],
            )


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for entities.container"""

import json
import unittest


from logilab.common import attrdict

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_seda.entities import (
    seda_profile_container_def,
    simplified_profile,
    full_seda2_profile,
    parent_and_container,
    rule_type_from_etype,
    custom,
    itree,
)

from cubicweb_seda.testutils import create_archive_unit, create_data_object


def sort_container(container_def):
    for k, v in container_def:
        yield k, sorted(v)


class ContainerTC(CubicWebTC):
    def test_seda_profile_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = seda_profile_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(
            container_def["SEDAMimeType"], [("seda_mime_type_from", "subject")]
        )
        self.assertNotIn("ConceptScheme", container_def)
        self.assertNotIn("Concept", container_def)
        self.assertNotIn("AuthorityRecord", container_def)
        entity = self.vreg["etypes"].etype_class("SEDAArchiveTransfer")(self)
        self.assertIsNotNone(entity.cw_adapt_to("IContainer"))
        self.assertIsNone(entity.cw_adapt_to("IContained"))

    def test_container_relation(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            transfer = create("SEDAArchiveTransfer", title="test profile")
            mtclv = create(
                "SEDAMimeTypeCodeListVersion",
                seda_mime_type_code_list_version_from=transfer,
            )
            access_rule = create("SEDAAccessRule", seda_access_rule=transfer)
            cnx.commit()
            for entity in (mtclv, access_rule):
                entity.cw_clear_all_caches()
                self.assertEqual(
                    entity.cw_adapt_to("IContained").container.eid, transfer.eid
                )
            access_rule_seq = create(
                "SEDASeqAccessRuleRule", reverse_seda_seq_access_rule_rule=access_rule
            )
            start_date = create("SEDAStartDate", seda_start_date=access_rule_seq)
            cnx.commit()
            for entity in (access_rule_seq, start_date):
                entity.cw_clear_all_caches()
                self.assertEqual(
                    entity.cw_adapt_to("IContained").container.eid, transfer.eid
                )

    def test_container_relation_archive_unit(self):
        """Ensure that container relation is deleted upon deletion on
        "seda_archive_unit" relation.
        """
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test")
            unit, _, _ = create_archive_unit(
                transfer, cnx=cnx, user_cardinality="1", user_annotation="plop"
            )
            cnx.commit()
            unit.cw_clear_all_caches()
            self.assertEqual(unit.container, (transfer,))
            self.assertIn(unit, transfer.reverse_seda_archive_unit)
            transfer.cw_set(reverse_seda_archive_unit=None)
            cnx.commit()
            unit.cw_clear_all_caches()
            self.assertEqual(unit.container, ())

    def test_archive_unit_container_clone(self):
        """Functional test for SEDA component clone."""
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = create_archive_unit(None, cnx=cnx)
            bdo = create_data_object(unit_alt_seq)
            cnx.commit()

            unit.cw_clear_all_caches()
            self.assertEqual(unit.container, ())  # XXX arguable
            unit_alt_seq.cw_clear_all_caches()
            self.assertEqual(unit_alt_seq.container[0].eid, unit.eid)
            bdo.cw_clear_all_caches()
            self.assertEqual(bdo.container[0].eid, unit.eid)

            # test clone without reparenting
            cloned = cnx.create_entity(
                unit.cw_etype, user_annotation="clone", clone_of=unit
            )
            cnx.commit()
            cloned.cw_clear_all_caches()
            self.assertEqual(cloned.user_annotation, "clone")
            cloned_unit_alt_seq = cloned.first_level_choice.content_sequence
            self.assertEqual(cloned_unit_alt_seq.container[0].eid, cloned.eid)
            cloned_bdo = cloned_unit_alt_seq.reverse_seda_data_object_reference[
                0
            ].seda_data_object_reference_id[0]
            self.assertEqual(cloned_bdo.container[0].eid, cloned.eid)

            # test clone with reparenting
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            cloned = cnx.create_entity(
                unit.cw_etype,
                user_annotation="I am mandatory",
                clone_of=unit,
                seda_archive_unit=transfer,
            )
            cnx.commit()
            cloned.cw_clear_all_caches()
            self.assertEqual(cloned.container[0].eid, transfer.eid)
            cloned.cw_clear_all_caches()
            cloned_unit_alt_seq = cloned.first_level_choice.content_sequence
            self.assertEqual(cloned_unit_alt_seq.container[0].eid, transfer.eid)
            cloned_bdo = cloned_unit_alt_seq.reverse_seda_data_object_reference[
                0
            ].seda_data_object_reference_id[0]
            self.assertEqual(cloned_bdo.container[0].eid, transfer.eid)
            cloned_bdo.cw_clear_all_caches()
            self.assertEqual(cloned_bdo.seda_binary_data_object[0].eid, transfer.eid)

    def test_archive_transfer_clone(self):
        """Functional test for SEDA profile cloning."""
        with self.admin_access.cnx() as cnx:
            scheme = cnx.create_entity("ConceptScheme", title="Algorithms")
            concept = scheme.add_concept("md5")
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            cnx.create_entity(
                "SEDAMessageDigestAlgorithmCodeListVersion",
                seda_message_digest_algorithm_code_list_version_from=transfer,
                seda_message_digest_algorithm_code_list_version_to=scheme,
            )
            unit, unit_alt, unit_alt_seq = create_archive_unit(
                transfer, cnx=cnx, title="hello"
            )
            create_data_object(transfer, seda_algorithm=concept)
            cnx.commit()

            clone = cnx.create_entity(
                "SEDAArchiveTransfer", title="Clone", clone_of=transfer
            )
            cnx.commit()

            self.assertEqual(
                (
                    clone.reverse_seda_message_digest_algorithm_code_list_version_from[
                        0
                    ]
                    .seda_message_digest_algorithm_code_list_version_to[0]
                    .eid
                ),
                scheme.eid,
            )
            seq = clone.archive_units[0].first_level_choice.content_sequence
            self.assertEqual(seq.title.title, "hello")
            self.assertEqual(
                transfer.binary_data_objects[0].seda_algorithm[0].eid, concept.eid
            )

    def test_container_clone_clone(self):
        """Functional test for SEDA component clone."""
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = create_archive_unit(None, cnx=cnx)
            cnx.commit()

            cloned = cnx.create_entity(
                unit.cw_etype, user_annotation="x", clone_of=unit
            )
            cnx.commit()
            cloned.cw_clear_all_caches()

            cloned2 = cnx.create_entity(unit.cw_etype, user_annotation="Y")
            # mimick what's occurs when you're using the copy form
            cloned2.copy_relations(cloned.eid)  # 1. copy relation
            self.assertFalse(cloned2.clone_of)
            cloned2.cw_set(
                clone_of=cloned.eid
            )  # 2. execute __linkto (which trigger cloning)

            cnx.commit()

            cloned2.cw_clear_all_caches()
            self.assertEqual(len(cloned2.seda_alt_archive_unit_archive_unit_ref_id), 1)

            unit_alt.cw_clear_all_caches()
            self.assertEqual(unit_alt.container[0].eid, unit.eid)
            self.assertEqual(
                cloned.seda_alt_archive_unit_archive_unit_ref_id[0].container[0].eid,
                cloned.eid,
            )
            self.assertEqual(
                cloned2.seda_alt_archive_unit_archive_unit_ref_id[0].container[0].eid,
                cloned2.eid,
            )

    def test_data_object_reference_clone(self):
        with self.admin_access.repo_cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            unit, unit_alt, unit_alt_seq = create_archive_unit(transfer)
            # don't add link from data object to transfer intentionally to force
            # going through the archive unit to clone it
            bdo = create_data_object(unit_alt_seq)
            cnx.commit()

            clone = cnx.create_entity(
                "SEDAArchiveTransfer", title="Clone", clone_of=transfer
            )
            cnx.commit()

            # Ensure data object is cloned through data_object_reference_id and
            # container relation is properly handled
            bdo_clone = cnx.execute("Any MAX(X) WHERE X is SEDABinaryDataObject").one()
            self.assertNotEqual(bdo_clone.eid, bdo.eid)
            bdo.cw_clear_all_caches()
            self.assertEqual([e.eid for e in bdo.container], [transfer.eid])
            self.assertEqual([e.eid for e in bdo_clone.container], [clone.eid])

            self.assertFalse(
                cnx.execute("Any X GROUPBY X WHERE X container C HAVING COUNT(C) > 1")
            )
            self.assertFalse(cnx.execute("Any X WHERE NOT X container C"))


class FakeEntity:
    cw_etype = "Whatever"

    def __init__(self, _cw):
        self._cw = _cw

    def has_eid(self):
        return False


class PredicatesTC(CubicWebTC):
    def test_simplified_profile(self):
        simplified_profile_pred = simplified_profile()
        full_seda2_profile_pred = full_seda2_profile()
        with self.admin_access.web_request() as req:
            transfer = req.create_entity("SEDAArchiveTransfer", title="test profile")
            self.assertEqual(simplified_profile_pred(None, req, entity=transfer), 0)
            self.assertEqual(full_seda2_profile_pred(None, req, entity=transfer), 1)
            self.assertEqual(
                simplified_profile_pred(None, req, rset=transfer.as_rset()), 0
            )
            self.assertEqual(
                full_seda2_profile_pred(None, req, rset=transfer.as_rset()), 1
            )
            access_rule = req.create_entity("SEDAAccessRule", seda_access_rule=transfer)
            req.cnx.commit()  # needed to set the container relation
            self.assertEqual(simplified_profile_pred(None, req, entity=access_rule), 0)
            self.assertEqual(full_seda2_profile_pred(None, req, entity=access_rule), 1)
            req.form["__linkto"] = f"whatever:{transfer.eid}:whatever"
            being_created = FakeEntity(req)
            self.assertEqual(
                simplified_profile_pred(None, req, entity=being_created), 0
            )
            self.assertEqual(
                full_seda2_profile_pred(None, req, entity=being_created), 1
            )
            del req.form["__linkto"]
            req.form["arg"] = [str(transfer.eid)]
            self.assertEqual(
                simplified_profile_pred(None, req, entity=being_created), 0
            )
            self.assertEqual(
                full_seda2_profile_pred(None, req, entity=being_created), 1
            )
            transfer.cw_set(simplified_profile=True)
            self.assertEqual(simplified_profile_pred(None, req, entity=transfer), 1)
            self.assertEqual(full_seda2_profile_pred(None, req, entity=transfer), 0)

            req.form = {"etype": "SEDAArchiveUnit"}
            etype_vreg = req.vreg["etypes"]
            unit = etype_vreg.etype_class("SEDAArchiveUnit")(req)
            self.assertEqual(simplified_profile_pred(None, req, entity=unit), 1)
            entity = etype_vreg.etype_class(
                "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement"
            )(req)
            self.assertEqual(simplified_profile_pred(None, req, entity=entity), 1)


class ITreeTC(CubicWebTC):
    def assertChildren(self, entity, expected_eids):
        entity.cw_clear_all_caches()
        itree = entity.cw_adapt_to("ITreeBase")
        children = [x.eid for x in itree.iterchildren()]
        self.assertEqual(children, expected_eids)

    def assertParent(self, entity, expected_eid):
        entity.cw_clear_all_caches()
        itree = entity.cw_adapt_to("ITreeBase")
        parent = itree.parent()
        if parent:
            parent_eid = parent.eid
        else:
            parent_eid = None
        self.assertEqual(parent_eid, expected_eid)

    def test(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            au, alt, seq = create_archive_unit(transfer)
            au2 = create_archive_unit(transfer)[0]
            do_ref = cnx.create_entity(
                "SEDADataObjectReference", seda_data_object_reference=seq
            )
            bdo = create_data_object(
                transfer, reverse_seda_data_object_reference_id=do_ref
            )
            do_ref2 = cnx.create_entity(
                "SEDADataObjectReference", seda_data_object_reference=seq
            )
            bdo2 = create_data_object(
                transfer, reverse_seda_data_object_reference_id=do_ref2
            )
            cnx.commit()

            self.assertChildren(transfer, [bdo.eid, bdo2.eid, au.eid, au2.eid])
            self.assertChildren(au, [])
            self.assertParent(transfer, None)
            self.assertParent(au, transfer.eid)
            self.assertParent(bdo, transfer.eid)

            transfer.cw_set(simplified_profile=True)
            cnx.commit()

            self.assertChildren(transfer, [au.eid, au2.eid])
            self.assertChildren(au, [bdo.eid, bdo2.eid])
            self.assertParent(transfer, None)
            self.assertParent(au, transfer.eid)
            self.assertParent(bdo, au.eid)

            au.cw_set(ordering=2)
            au2.cw_set(ordering=1)
            bdo.cw_set(ordering=2)
            bdo2.cw_set(ordering=1)
            do_ref.cw_set(ordering=2)
            do_ref2.cw_set(ordering=1)
            cnx.commit()

            self.assertChildren(transfer, [au2.eid, au.eid])
            self.assertChildren(au, [bdo2.eid, bdo.eid])

            transfer.cw_set(simplified_profile=False)
            cnx.commit()

            self.assertChildren(transfer, [bdo2.eid, bdo.eid, au2.eid, au.eid])


class ReorderTC(CubicWebTC):
    def test_reorder_base(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="Test")
            bdo1 = create_data_object(transfer)
            bdo2 = create_data_object(transfer)
            bdo3 = create_data_object(transfer)
            cnx.commit()

            itree.move_child_at_index(cnx, transfer.eid, bdo3.eid, 0)
            transfer.cw_clear_all_caches()
            self.assertEqual(
                [
                    (x.eid, x.ordering)
                    for x in transfer.cw_adapt_to("ITreeBase").iterchildren()
                ],
                [(bdo3.eid, 1), (bdo1.eid, 2), (bdo2.eid, 3)],
            )

            itree.move_child_at_index(cnx, transfer.eid, bdo3.eid, 3)
            transfer.cw_clear_all_caches()
            self.assertEqual(
                [
                    (x.eid, x.ordering)
                    for x in transfer.cw_adapt_to("ITreeBase").iterchildren()
                ],
                [(bdo1.eid, 1), (bdo2.eid, 2), (bdo3.eid, 3)],
            )

            itree.move_child_at_index(cnx, transfer.eid, bdo1.eid, 2)
            transfer.cw_clear_all_caches()
            self.assertEqual(
                [
                    (x.eid, x.ordering)
                    for x in transfer.cw_adapt_to("ITreeBase").iterchildren()
                ],
                [(bdo2.eid, 1), (bdo1.eid, 2), (bdo3.eid, 3)],
            )

            itree.move_child_at_index(cnx, transfer.eid, bdo3.eid, 1)
            transfer.cw_clear_all_caches()
            self.assertEqual(
                [
                    (x.eid, x.ordering)
                    for x in transfer.cw_adapt_to("ITreeBase").iterchildren()
                ],
                [(bdo2.eid, 1), (bdo3.eid, 2), (bdo1.eid, 3)],
            )


class RuleFromETypeTC(unittest.TestCase):
    def test_rule_from_etype(self):
        for rule_type in (
            "access",
            "appraisal",
            "classification",
            "reuse",
            "dissemination",
            "storage",
        ):
            for prefix, suffix in [
                ("SEDAAlt", "RulePreventInheritance"),
                ("SEDASeq", "RuleRule"),
                ("SEDA", "Rule"),
            ]:
                self.assertEqual(
                    rule_type_from_etype(prefix + rule_type.capitalize() + suffix),
                    rule_type,
                )


class ParentAndContainerTC(CubicWebTC):
    def test_nodata(self):
        with self.admin_access.web_request() as req:
            parent, container = parent_and_container(
                attrdict(_cw=req, has_eid=lambda: False)
            )
            self.assertIsNone(parent)
            self.assertIsNone(container)

    def test_linkto(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity(
                "SEDAArchiveTransfer", title="test profile"
            )
            req.form["__linkto"] = f"x:{transfer.eid}:y"
            parent, container = parent_and_container(
                attrdict(_cw=req, has_eid=lambda: False)
            )
            self.assertEqual(parent.eid, transfer.eid)
            self.assertEqual(container.eid, transfer.eid)

    def test_arg(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity(
                "SEDAArchiveTransfer", title="test profile"
            )
            req.form["arg"] = [json.dumps(transfer.eid)]
            parent, container = parent_and_container(
                attrdict(_cw=req, has_eid=lambda: False)
            )
            self.assertEqual(parent.eid, transfer.eid)
            self.assertEqual(container.eid, transfer.eid)

    def test_eid(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity(
                "SEDAArchiveTransfer", title="test profile"
            )
            req.form["sedaContainerEID"] = str(transfer.eid)
            parent, container = parent_and_container(
                attrdict(_cw=req, has_eid=lambda: False)
            )
            self.assertIsNone(parent)
            self.assertEqual(container.eid, transfer.eid)


class CustomEntitiesTC(CubicWebTC):
    def test_title(self):
        with self.admin_access.cnx() as cnx:
            for etype in (
                "SEDAArchiveUnit",
                "SEDABinaryDataObject",
                "SEDAPhysicalDataObject",
            ):
                ent = cnx.create_entity(etype, user_annotation="bla bla\nbli bli blo\n")
                self.assertEqual(ent.dc_title(), "bla bla")

    def test_climb_rule_holders(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            unit, _, unit_alt_seq = create_archive_unit(transfer)
            subunit, _, subunit_alt_seq = create_archive_unit(unit_alt_seq)
            cnx.commit()

            self.assertEqual(
                list(custom._climb_rule_holders(subunit_alt_seq)),
                [subunit_alt_seq, unit_alt_seq, transfer],
            )
            self.assertEqual(
                list(custom._climb_rule_holders(subunit)),
                [subunit_alt_seq, unit_alt_seq, transfer],
            )
            self.assertEqual(
                list(custom._climb_rule_holders(unit_alt_seq)), [unit_alt_seq, transfer]
            )
            self.assertEqual(
                list(custom._climb_rule_holders(unit)), [unit_alt_seq, transfer]
            )
            self.assertEqual(list(custom._climb_rule_holders(transfer)), [transfer])


if __name__ == "__main__":
    unittest.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Generate CubicWeb's uicfg rules from XSD file.

XSD parsing is done using generateDS, which has been copied into the `gends` directory (only the
used bits).
"""


from cubicweb import neg_role, _

from cubicweb_seda.xsd import XSDM_MAPPING
from cubicweb_seda.xsd2yams import CodeGenerator


FIRST_LEVEL_ETYPES = {
    "SEDAArchiveTransfer",
    "SEDABinaryDataObject",
    "SEDAPhysicalDataObject",
    "SEDAArchiveUnit",
}

RTYPES_IN_TAB = {
    "seda_binary_data_object",
    "seda_physical_data_object",
    "seda_related_transfer_reference",
    "seda_archive_unit",
    "seda_relationship",
    "seda_storage_rule",
    "seda_appraisal_rule",
    "seda_access_rule",
    "seda_dissemination_rule",
    "seda_reuse_rule",
    "seda_classification_rule",
    "seda_need_authorization",
    "seda_restriction_rule_id_ref",
    "seda_restriction_value",
    "seda_restriction_end_date",
}
for element_name in (
    "CodeListVersions",
    "FormatIdentification",
    "FileInfo",
    "PhysicalDimensions",
    "Gps",
    "RelatedObjectReference",
    "CustodialHistory",
    "Coverage",
):
    for rtype, role, path in XSDM_MAPPING.iter_rtype_role(element_name):
        RTYPES_IN_TAB.add(rtype)


class UICFGGenerator(CodeGenerator):
    """UICFG rules generator"""

    rtags_info = {
        "actionbox_appearsin_addmenu": {
            "shortname": "abaa",
            "subject": "abaa.tag_subject_of(('*', '{rtype}', '*'), {value})",
            "object": "abaa.tag_object_of(('*', '{rtype}', '*'), {value})",
        },
        "autoform_section": {
            "shortname": "afs",
            "subject": "afs.tag_subject_of(('*', '{rtype}', '*'), 'main', '{value}')",
            "object": "afs.tag_object_of(('*', '{rtype}', '*'), 'main', '{value}')",
        },
        "autoform_field_kwargs": {
            "shortname": "affk",
            "subject": "affk.tag_subject_of(('*', '{rtype}', '*'), {value})",
            "object": "affk.tag_object_of(('*', '{rtype}', '*'), {value})",
        },
        "primaryview_section": {
            "shortname": "pvs",
            "subject": "pvs.tag_subject_of(('*', '{rtype}', '*'), '{value}')",
            "object": "pvs.tag_object_of(('*', '{rtype}', '*'), '{value}')",
        },
        "reledit_ctrl": {
            "shortname": "rec",
            "subject": "rec.tag_subject_of(('*', '{rtype}', '*'), "
            "{{'novalue_label': '{value}'}})",
            "object": "rec.tag_object_of(('*', '{rtype}', '*'), "
            "{{'novalue_label': '{value}'}})",
        },
    }

    def _generate(self, mapping, stream):
        self._processed = set()
        stream.write("from cubicweb.web import formwidgets as fw\n")
        stream.write("from cubicweb.web.views import uicfg\n\n")
        # indexview_etype_section configuration
        stream.write("ives = uicfg.indexview_etype_section\n")
        all_etypes = set()
        for mapping_element in mapping.ordered:
            etypes = self._callback("etypes_for", mapping_element)
            all_etypes.update(set(etypes))
        all_etypes.remove("SEDAArchiveTransfer")
        for etype in all_etypes:
            stream.write(f"ives['{etype}'] = 'subobject'\n")
        # autoform / primary view relation rtags
        for rtag, rtag_info in sorted(self.rtags_info.items()):
            alias = rtag_info["shortname"]
            stream.write(f"\n\n{rtag_info['shortname']} = uicfg.{rtag}\n")
            for mapping_element in mapping.ordered:
                for rtype, role, value in self._callback(
                    rtag + "_for", mapping_element
                ):
                    template = rtag_info[role]
                    stream.write(template.format(**locals()) + "\n")
        # fields order
        stream.write("pvds = uicfg.primaryview_display_ctrl\n")
        for mapping_element in mapping.ordered:
            for etype, attributes in self._callback("order_for", mapping_element):
                stream.write(f"affk.set_fields_order('{etype}', {attributes})\n")
                stream.write(f"pvds.set_fields_order('{etype}', {attributes})\n")
        # fields documentation
        stream.write("\nETYPE_ATTR_DOC = {}\n")
        template = (
            "ETYPE_ATTR_DOC[('{etype}', '{rtype}', '{role}')] = ('{element}', {doc})\n"
        )
        for mapping_element in mapping.ordered:
            for etype, rtype, role, element, doc in self._callback(
                "doc_for", mapping_element
            ):
                stream.write(
                    template.format(
                        etype=etype, rtype=rtype, role=role, element=element, doc=doc
                    )
                )

    def etypes_for_e_type_mapping(self, mapping):
        yield mapping.etype

    def autoform_section_for_rdef_mapping(self, mapping):
        if ("afs", mapping.rtype) in self._processed:
            return
        self._processed.add(("afs", mapping.rtype))
        if mapping.rtype in RTYPES_IN_TAB:
            section = "hidden"
            role = mapping.composite or "subject"
        elif mapping.composite:
            section = "inlined"
            role = mapping.composite
        else:
            section = "attributes"
            role = "subject"
        yield mapping.rtype, neg_role(role), "hidden"
        yield mapping.rtype, role, section

    def autoform_field_kwargs_for_e_type_mapping(self, mapping):
        for rtype, target_etype in sorted(mapping.attributes.items()):
            if rtype == "id":
                continue
            if target_etype == "String" and ("affk", rtype) not in self._processed:
                self._processed.add(("affk", rtype))
                yield rtype, "subject", {"widget": Code("fw.TextInput({'size': 80})")}
            elif target_etype == "Boolean" and ("aff", rtype) not in self._processed:
                self._processed.add(("aff", rtype))
                yield rtype, "subject", {"allow_none": True}

    def autoform_field_kwargs_for_rdef_mapping(self, mapping):
        if mapping.rtype.endswith("code_list_version_to"):
            yield mapping.rtype, "subject", {"label": "value"}

    def primaryview_section_for_rdef_mapping(self, mapping):
        if ("pvs", mapping.rtype) in self._processed:
            return
        self._processed.add(("pvs", mapping.rtype))
        if mapping.rtype in RTYPES_IN_TAB:
            section = "hidden"
            role = mapping.composite or "subject"
            yield mapping.rtype, neg_role(role), section
            yield mapping.rtype, role, section
        elif (
            "Concept" in mapping.objtypes
            or "ConceptScheme" in mapping.objtypes
            or "AuthorityRecord" in mapping.objtypes
        ):
            yield mapping.rtype, "object", "hidden"

    def reledit_ctrl_for_e_type_mapping(self, mapping):
        for rtype, target_etype in sorted(mapping.attributes.items()):
            if rtype == "id":
                continue
            if ("rec", rtype) not in self._processed:
                self._processed.add(("rec", rtype))
                yield rtype, "subject", _("<no value specified>")

    def reledit_ctrl_for_rdef_mapping(self, mapping):
        if ("rec", mapping.rtype) in self._processed:
            return
        self._processed.add(("rec", mapping.rtype))
        if mapping.composite is None:
            yield mapping.rtype, "subject", "<no value specified>"
        else:
            card = mapping.card[mapping.composite == "object"]
            if card == "1":
                yield mapping.rtype, mapping.composite, _("<no value specified>")
            elif card == "?":
                yield mapping.rtype, mapping.composite, " "

    def actionbox_appearsin_addmenu_for_rdef_mapping(self, mapping):
        if ("abaa", mapping.rtype) in self._processed:
            return
        self._processed.add(("abaa", mapping.rtype))
        role = mapping.composite or "subject"
        yield mapping.rtype, neg_role(role), False
        yield mapping.rtype, role, False

    def order_for_e_type_mapping(self, mapping):
        attributes = ordered_attributes(mapping)
        if len(attributes) > 1:
            yield mapping.etype, attributes

    def doc_for_rdef_mapping(self, mapping):
        if mapping.element_name or mapping.desc:
            rtype = mapping.rtype
            yield mapping.subjtype, rtype, "subject", mapping.element_name, mapping.desc
            for etype in mapping.objtypes:
                yield etype, rtype, "object", mapping.element_name, mapping.desc


class Code(str):
    """Special string subclass whose repr() doesn't add quotes, for insertion of python code in a
    data structure
    """

    def __repr__(self):
        return str(self)


def ordered_attributes(mapping):
    """Given an ETypeMapping, return a list of its attributes sorted by desired
    order of appearance"""
    attributes = [attr for attr in mapping.attributes if attr != "id"]
    if mapping.etype == "SEDAArchiveTransfer":
        attributes.append("title")
        attributes.append("user_annotation")
    elif mapping.etype in (
        "SEDAArchiveUnit",
        "SEDABinaryDataObject",
        "SEDAPhysicalDataObject",
    ):
        attributes.append("user_cardinality")
        attributes.append("user_annotation")
    elif attributes:
        if len(mapping.cards) > 1:
            attributes.insert(0, "user_cardinality")
            attributes.append("user_annotation")
    return attributes


if __name__ == "__main__":
    UICFGGenerator.main()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda adapter classes for generation of a profile generation as HTML"""

from cubicweb import _

from .profile_generation import SEDA2ExportAdapter, content_types
from .profile_generation import xselement_scheme_attribute, _concept_value


def element_uml_cardinality(occ, card_entity):
    """Return UML like cardinality for the given pyxst Occurence. Cardinality may be
    overriden by the data model's user_cardinality value.
    """
    cardinality = getattr(card_entity, "user_cardinality", None)
    if cardinality is None:
        minimum = occ.minimum
        maximum = occ.maximum
        if minimum == maximum == 1:
            return "1"
        elif maximum != 1:
            maximum = "n"
        return f"{minimum}..{maximum}"
    else:
        return cardinality


def attribute_cardinality_as_string(occ, card_entity):
    """Return 'optional' or 'mandatory' for the given pyxst attribute's
    Occurence. Cardinality may be overriden by the data model's user_cardinality
    value."""
    cardinality = getattr(card_entity, "user_cardinality", None)
    if cardinality is None:
        minimum = occ.minimum
    else:
        # XXX assert cardinality in ('0..1', '1'), cardinality
        if cardinality[0] == "0":
            minimum = 0
        else:
            minimum = 1
    return _("mandatory") if minimum else _("optional")


class SEDA2HTMLExport(SEDA2ExportAdapter):
    """Adapter to build a HTML representation of a SEDA profile"""

    __regid__ = "SEDA-2.1.html"
    namespaces = {}
    css = """
    h1{
      font-variant: small-caps;
      color : #FF9800;
      text-align: center;
      padding : 10px;
      font-weight: lighter;
      border-bottom : 2px solid  #E0E0E0;
    }

    a{
      color : #EF6C00;
    }
    body {
      font-family: helvetica, Arial;
      background-color: #FAFAFA;
      color : #666;
    }
    body >div{
      background-color: #FFF;
      padding-bottom : 20px;
      margin-top : 20px;
      box-shadow: 2px 2px 2px 0px #E0E0E0;
    }

    body >div >h3{
      font-variant: normal;
      font-weight: lighter;
      font-size:15px;
      padding : 10px;
      background-color: #EF6C00;
      color : white;
      font-variant: small-caps;
      cursor : pointer;
    }

    body >div >h3 span.card{
      font-size: 70%;
      color : #E0E0E0;
      margin-left : 5px;
    }

    body>div>div, .sequence >div{
      margin-top : 20px
    }

    body>div>div div{
      padding-left : 20px;
      margin-left: 15px;
      border-left : 1px dashed #EF6C00;
    }

    body>div>div h3{
      margin-bottom : 10px;
    }

    h3 span.card {
      font-size: 70%;
      color : #EF6C00;
      margin-left : 5px;
    }

    span {
      font-size: 90%;
      color: #555;
    }
    span.label {
      font-weight: bold;
      padding-left: 1px;
    }

    div.attribute {
      color: #888;
      width: 40em;
      font-size: small;
      padding : 10px;
    }

    div.attribute span.label {
      float: left;
      color : #EF6C00;
      font-weight: lighter;
      margin-right : 20px;
    }

    .sequence >h3:first-child{
      color :  #EF6C00;
    }
    """

    def dump_etree(self):
        """Return an XSD etree for the adapted SEDA profile."""
        root = self.element("html")
        head = self.element("head", root)
        self.element("title", head, text=self.entity.dc_title())
        self.element(
            "meta",
            head,
            {"http-equiv": "content-type", "content": "text/html; charset=UTF-8"},
        )
        self.element("style", head, text=self.css)
        body = self.element("body", root)
        self._dump(body)
        return root

    def init_transfer_element(self, xselement, root, entity):
        self.element("h1", root, text=entity.title)
        if entity.user_annotation:
            self.element(
                "div",
                root,
                text=entity.user_annotation,
                attributes={"class": "description"},
            )
        return root

    def jumped_element(self, profile_element):
        return profile_element[-1]

    def element_alternative(
        self, occ, profile_element, target_value, to_process, card_entity
    ):
        div = self.element("div", profile_element, attributes={"class": "choice"})
        self.title(div, self._cw._("Alternative"), occ, card_entity)
        to_process[occ.target].append((target_value, div))

    def element_sequence(
        self, occ, profile_element, target_value, to_process, card_entity
    ):
        div = self.element("div", profile_element, attributes={"class": "sequence"})
        self.title(div, self._cw._("Sequence"), occ, card_entity)
        to_process[occ.target].append((target_value, div))

    def element_xmlattribute(
        self, occ, profile_element, target_value, to_process, card_entity
    ):
        div = self.element("div", profile_element, attributes={"class": "attribute"})
        card = self._cw._(attribute_cardinality_as_string(occ, card_entity))
        self.element(
            "span", div, text=occ.target.local_name, attributes={"class": "label"}
        )
        self.element("span", div, text=card, attributes={"class": "card"})
        fixed_value = self.serialize(target_value)
        if isinstance(fixed_value, str):
            self.element("span", div, text=fixed_value, attributes={"class": "value"})
        else:
            span = self.element("span", div, attributes={"class": "value"})
            if fixed_value is not None:
                span.append(fixed_value)

    def element_xmlelement(
        self, occ, profile_element, target_value, to_process, card_entity
    ):
        xselement = occ.target
        if isinstance(occ, dict):  # fake occurence introduced for some elements'content
            return
        attrs = {}
        if hasattr(target_value, "id"):
            attrs["id"] = target_value.id
        div = self.element("div", profile_element, attrs)
        self.title(div, xselement.local_name, occ, card_entity)
        annotation = getattr(card_entity, "user_annotation", None)
        if annotation:
            self.element(
                "div", div, text=annotation, attributes={"class": "description"}
            )
        xstypes = content_types(xselement.textual_content_type)
        self.fill_element(xselement, div, target_value, card_entity, xstypes)
        if getattr(target_value, "eid", None):  # value is an entity
            to_process[xselement].append((target_value, div))

    def fill_element(self, xselement, div, target_value, card_entity, xstypes=None):
        if xselement.local_name == "KeywordType":
            list_div = self.element("div", div)
            self.element(
                "span", list_div, text="listVersionID", attributes={"class": "label"}
            )
            if target_value:
                list_value = (
                    target_value.scheme.description or target_value.scheme.dc_title()
                )
            else:
                list_value = "edition 2009"
            self.element(
                "span", list_div, text=list_value, attributes={"class": "value"}
            )

        elif xselement.local_name == "KeywordReference" and card_entity.scheme:
            self.concept_scheme_attribute(xselement, div, card_entity.scheme)

        elif getattr(target_value, "cw_etype", None) == "Concept":
            self.concept_scheme_attribute(xselement, div, target_value.scheme)
        if xstypes:
            ct_div = self.element("div", div)
            self.element(
                "span",
                ct_div,
                text=self._cw._("XSD content type"),
                attributes={"class": "label"},
            )
            xstypes = self._cw._(" ALT_I18N ").join(
                "xsd:" + xstype for xstype in xstypes
            )
            self.element("span", ct_div, text=xstypes, attributes={"class": "value"})
        if target_value is None:
            values = ()
        elif not isinstance(target_value, (tuple, list)):
            values = [target_value]
        else:
            values = target_value
        value_div = None
        for value in values:
            fixed_value = self.serialize(value)
            if fixed_value is None:
                continue
            if value_div is None:
                value_div = self.element("div", div)
                self.element(
                    "span",
                    value_div,
                    text=self._cw._("fixed value"),
                    attributes={"class": "label"},
                )
            if isinstance(fixed_value, str):
                self.element(
                    "span", value_div, text=fixed_value, attributes={"class": "value"}
                )
            else:
                span = self.element("span", value_div, attributes={"class": "value"})
                span.append(fixed_value)

    def concept_scheme_attribute(self, xselement, div, scheme):
        try:
            xsattr = xselement_scheme_attribute(xselement)
        except KeyError:
            return  # no attribute to define scheme (e.g. for language specification)
        div = self.element("div", div)
        self.element("span", div, text=xsattr, attributes={"class": "label"})
        self.element(
            "a",
            self.element("span", div, attributes={"class": "value"}),
            {"href": scheme.absolute_url()},
            text=scheme.dc_title(),
        )

    def title(self, div, label, occ, card_entity):
        h3 = self.element("h3", div, text=label)
        self.element(
            "span",
            h3,
            attributes={"class": "card"},
            text=f"[{self._cw._(element_uml_cardinality(occ, card_entity))}]",
        )

    def serialize(self, value):
        """Return value as None or some text or etree node to be inserted in the HTML DOM."""
        if value is None:
            return None
        if hasattr(value, "eid"):
            if value.cw_etype in ("AuthorityRecord", "ConceptScheme"):
                href, label = value.absolute_url(), value.dc_title()
            elif value.cw_etype == "Concept":
                href = value.absolute_url()
                label = _concept_value(value, "seda-2")
                if value.label() != label:
                    label = f"{label} ({value.label()})"
            elif hasattr(value, "id"):
                # value is something in the profile which has a id (e.g. archive unit, data object)
                href, label = f"#{value.id}", value.id
            else:
                return None  # intermediary entity
            return self.element(
                "a",
                self.element("span", attributes={"class": "value"}),
                attributes={"href": href},
                text=label,
            )
        else:
            if isinstance(value, bool):
                value = "true" if value else "false"
            assert isinstance(value, str), repr(value)
            return value



sync_schema_props_perms("language_code")

scheme = cnx.find("ConceptScheme", title="Langues (ISO-639-3)").one()
with cnx.deny_all_hooks_but("metadata"):
    rset = cnx.execute(
        'Any C, L, LL WHERE L label_of C, L language_code "seda-2", L label LL,'
        "C in_scheme CS, CS eid %(cs)s",
        {"cs": scheme.eid},
    )
    for i, concept in enumerate(rset.entities()):
        label = rset.get_entity(i, 1)
        label.cw_set(language_code="seda")
        cnx.create_entity(
            "Label", label_of=concept, language_code="seda-02", label=label.label[:2]
        )
    commit()



from cubicweb_seda.dataimport import LCSV_FILES, import_seda_schemes

# please close your eyes - this is necessary if seda is used by saem else call
# to import_seda_schemes will fail because it will attempt to reuse and existing
# uri. Should be done there but still easier here.
try:
    from cubicweb_saem_ref.site_cubicweb import init_seda_scheme

    init_seda_scheme.__defaults__[0][0] = len(LCSV_FILES) - 1
except ImportError:
    pass

print("filtering language concepts, this will take a while")

seda_02_langs = {
    "aa",
    "ab",
    "af",
    "ak",
    "am",
    "ar",
    "an",
    "as",
    "av",
    "ae",
    "ay",
    "az",
    "ba",
    "bm",
    "be",
    "bn",
    "bh",
    "bi",
    "bo",
    "bs",
    "br",
    "bg",
    "ca",
    "cs",
    "ch",
    "ce",
    "cu",
    "cv",
    "co",
    "cr",
    "cy",
    "da",
    "de",
    "dv",
    "dz",
    "el",
    "en",
    "eo",
    "et",
    "ee",
    "es",
    "eu",
    "fo",
    "fa",
    "fj",
    "fi",
    "fr",
    "fy",
    "ff",
    "ga",
    "gl",
    "gv",
    "gn",
    "gu",
    "ht",
    "ha",
    "he",
    "hz",
    "hi",
    "ho",
    "hr",
    "hu",
    "hy",
    "ig",
    "is",
    "io",
    "ii",
    "iu",
    "ie",
    "ia",
    "id",
    "ik",
    "it",
    "jv",
    "ja",
    "ka",
    "kl",
    "kn",
    "ks",
    "kr",
    "kk",
    "km",
    "ki",
    "ky",
    "kv",
    "kg",
    "ko",
    "kj",
    "ku",
    "kw",
    "lo",
    "lv",
    "li",
    "ln",
    "lt",
    "lb",
    "lu",
    "lg",
    "mk",
    "mh",
    "ml",
    "mi",
    "mr",
    "ms",
    "mg",
    "mt",
    "mn",
    "my",
    "na",
    "nv",
    "nr",
    "nd",
    "ng",
    "ne",
    "nn",
    "nb",
    "no",
    "ny",
    "nl",
    "oc",
    "oj",
    "or",
    "om",
    "os",
    "pa",
    "pi",
    "pl",
    "pt",
    "ps" "qu",
    "rm",
    "ro",
    "rn",
    "ru",
    "rw",
    "sg",
    "sa",
    "si",
    "sk",
    "sl",
    "se",
    "sm",
    "sn",
    "sd",
    "so",
    "st",
    "sc",
    "sr",
    "ss",
    "su",
    "sw",
    "sv",
    "sq",
    "ty",
    "ta",
    "tt",
    "te",
    "tg",
    "tl",
    "th",
    "ti",
    "to",
    "tn",
    "ts",
    "tk",
    "tr",
    "tw",
    "ug",
    "uk",
    "ur",
    "uz",
    "ve",
    "vi",
    "vo",
    "wa",
    "wo",
    "xh",
    "yi",
    "yo",
    "za",
    "zu",
    "zh",
}

rset = rql(
    'Any X WHERE X in_scheme S, S title "Langues (ISO-639-3)", '
    'L label_of X, L language_code "seda-02", NOT L label IN ({})'.format(
        ",".join(repr(code) for code in seda_02_langs)
    )
)
print(len(rset), "concepts to delete")
for i, concept in enumerate(rset.entities()):
    concept.cw_delete()
    if i % 10:
        cnx.commit()


sync_schema_props_perms("user_cardinality")

add_relation_type("file_category")

import_seda_schemes(cnx)

sync_schema_props_perms("seda_mime_type_to")
sync_schema_props_perms("seda_format_id_to")



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda views for data objects (BinaryDataObject / PhysicalDataObject)"""

import json


from logilab.mtconverter import xml_escape

from cubicweb import tags, view, _
from cubicweb.predicates import match_form_params, is_instance
from cubicweb_web.views import uicfg, tabs, ibreadcrumbs

from cubicweb_relationwidget import views as rwdg
from cubicweb_skos.views import widgets as skos

from ..xsd2yams import SCHEME_FROM_CONTAINER
from ..entities import parent_and_container, simplified_profile, full_seda2_profile
from . import rtags_from_xsd_element, add_subobject_link
from . import viewlib
from . import uicfg as sedauicfg  # noqa - ensure those rules are defined first


pvs = uicfg.primaryview_section
pvdc = uicfg.primaryview_display_ctrl
rec = uicfg.reledit_ctrl
afs = uicfg.autoform_section
affk = uicfg.autoform_field_kwargs


def scheme_rql_expr(container, etype, rtype):
    """Return RQL expression (right part of the WHERE) to retrieve the scheme (mapped as 'CS'
    variable) for some (entity type / relation type) couple, in the context of the given
    `container`.
    """
    try:
        scheme_from_container = SCHEME_FROM_CONTAINER[(etype, rtype)]
    except KeyError:
        scheme_from_container = SCHEME_FROM_CONTAINER[rtype]
    return scheme_from_container[container.cw_etype]


class ContainedRelationFacetWidget(rwdg.RelationFacetWidget):
    def _render_triggers(self, w, domid, form, field, rtype):
        req = form._cw
        parent, container = parent_and_container(form.edited_entity)
        assert container is not None
        rql_expr = scheme_rql_expr(container, form.edited_entity.cw_etype, field.name)
        if not req.execute("Any CS WHERE " + rql_expr, {"container": container.eid}):
            scheme_relations = [x for x in rql_expr.split() if x.startswith("seda_")]
            if not scheme_relations:
                w(
                    req._(
                        "there is no scheme available for this relation. "
                        "Contact the site administrator."
                    )
                )
                return
            if len(scheme_relations) == 1:
                scheme_relation = req._(scheme_relations[0])
            else:
                scheme_relation = req.__(scheme_relations[0] + "_object")
            w(
                req._("you must specify a scheme for {0} to select a value").format(
                    scheme_relation
                )
            )
        else:
            return super()._render_triggers(w, domid, form, field, rtype)

    def trigger_search_url(self, entity, url_params):
        """Overriden to add information about who is the container

        This information will be used later for proper vocabulary computation.
        """
        # first retrieve the container entity
        _, container = parent_and_container(entity)
        assert container is not None
        # and put it as an extra url param
        url_params["container"] = str(container.eid)
        return super().trigger_search_url(entity, url_params)


for key in SCHEME_FROM_CONTAINER:
    try:
        etype, rtype = key
    except ValueError:
        etype = "*"
        rtype = key
    affk.tag_subject_of(
        (etype, rtype, "*"),
        {"widget": ContainedRelationFacetWidget(dialog_options={"width": 800})},
    )


class ContainedSearchForRelatedEntitiesView(skos.SearchForRelatedConceptsView):
    __select__ = skos.SearchForRelatedConceptsView.__select__ & match_form_params(
        "container"
    )

    def constrained_rql(self):
        container = self._cw.entity_from_eid(int(self._cw.form["container"]))
        rql_expr = scheme_rql_expr(
            container, self.schema_rdef.subject, self.schema_rdef.rtype
        )
        return rql_expr + ", O in_scheme CS", {"container": container.eid}


pvs.tag_object_of(("*", "seda_data_object_reference_id", "*"), "relations")
pvdc.tag_object_of(
    ("*", "seda_data_object_reference_id", "*"),
    {
        "vid": "autolimited",
        "subvid": "seda.object-ref.archive-unit",
        "label": _("referenced by:"),
    },
)

pvs.tag_subject_of(("*", "seda_algorithm", "*"), "attributes")
pvs.tag_object_of(("*", "seda_target", "*"), "hidden")  # in the relationship tab
# hide file_category from the main tab, it lives in the format tab
pvs.tag_subject_of(("SEDABinaryDataObject", "file_category", "*"), "hidden")

for rtype in ("seda_compressed", "seda_data_object_version_from"):
    # hide relation from autoform because of limitation of _container_eid
    afs.tag_object_of(("*", rtype, "*"), "main", "hidden")
    pvs.tag_object_of(("*", rtype, "*"), "attributes")
    rec.tag_object_of(
        ("*", rtype, "*"), {"rvid": "seda.reledit.complexlink", "novalue_label": " "}
    )

rec.tag_subject_of(
    ("SEDABinaryDataObject", "seda_alt_binary_data_object_attachment", "*"),
    {"rvid": "seda.reledit.alternative", "novalue_label": " "},
)

rec.tag_object_of(
    ("*", "seda_compressed", "*"), {"rvid": "seda.reledit.text", "novalue_label": " "}
)


def uri_cardinality_vocabulary(form, field):
    req = form._cw
    if form.edited_entity.has_eid():
        parent_type = form.edited_entity.cw_adapt_to("IContained").parent.cw_etype
    else:
        try:
            # inlined creation form
            parent_type = json.loads(req.form["arg"][1])
        except KeyError:
            # edition through reledit
            parent_eid = req.form["eid"]
            parent_type = req.describe(int(parent_eid))[0]
    if parent_type in ("SEDABinaryDataObject", "SEDAAltBinaryDataObjectAttachment"):
        return ["1"]
    return ["0..1", "1"]


affk.tag_attribute(
    ("SEDAUri", "user_cardinality"), {"choices": uri_cardinality_vocabulary}
)


bdo_ordered_fields = [
    ("user_cardinality", "subject"),
    ("user_annotation", "subject"),
    ("filename", "subject"),
    ("seda_date_created_by_application", "object"),
    ("seda_compressed", "object"),
    ("seda_data_object_version_from", "object"),
    ("seda_algorithm", "object"),
]
affk.set_fields_order("SEDABinaryDataObject", bdo_ordered_fields)
pvdc.set_fields_order("SEDABinaryDataObject", bdo_ordered_fields)


class BinaryDataObjectTabbedPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("SEDABinaryDataObject")
    tabs = [
        "main_tab",
        _("seda_bdo_format_identification"),
        _("seda_bdo_file_information"),
        _("seda_do_relations"),
    ]


def _setup_format_tab(rsection, display_ctrl):
    # hide format_id / mime_type, in favor of custom 'file_category'. Actual
    # format_id and mime_type will be derived from that value (XXX hook or profile
    # gen time ?)
    rsection.tag_object_of(
        ("*", "seda_format_id_from", "SEDABinaryDataObject"), "hidden"
    )
    rsection.tag_object_of(
        ("*", "seda_mime_type_from", "SEDABinaryDataObject"), "hidden"
    )
    rsection.tag_subject_of(
        ("SEDABinaryDataObject", "file_category", "*"), "attributes"
    )
    # dunno why it crashes without this
    display_ctrl.tag_subject_of(
        ("SEDABinaryDataObject", "file_category", "*"), {"order": 0}
    )


class BinaryDataObjectFormatIdentificationTab(viewlib.PrimaryTabWithoutBoxes):
    """Display format identification information of a binary data object"""

    __regid__ = "seda_bdo_format_identification"
    __select__ = is_instance("SEDABinaryDataObject")

    rsection, display_ctrl = rtags_from_xsd_element(
        "SEDABinaryDataObject", "FormatIdentification"
    )
    _setup_format_tab(rsection, display_ctrl)


class SimplifiedBinaryDataObjectFormatIdentificationTab(
    BinaryDataObjectFormatIdentificationTab
):
    __select__ = (
        BinaryDataObjectFormatIdentificationTab.__select__ & simplified_profile()
    )

    rsection, display_ctrl = rtags_from_xsd_element(
        "SEDABinaryDataObject", "FormatIdentification"
    )
    _setup_format_tab(rsection, display_ctrl)
    rsection.tag_object_of(
        ("*", "seda_format_litteral", "SEDABinaryDataObject"), "hidden"
    )


def file_category_vocabulary(form, field):
    rset = form._cw.execute(
        "Any C,CLL,E,ELL ORDERBY CLL,ELL WHERE "
        'C in_scheme CS, CS scheme_relation_type CR, CR name "file_category", '
        "NOT C broader_concept SC, E broader_concept C, "
        "CL label_of C, CL label CLL,"
        "EL label_of E, EL label ELL"
    )
    values = []
    current_category = None
    for category, category_label, extension, extension_label in rset:
        if current_category is None or current_category != category:
            current_category = category
            values.append((category_label, str(category)))
        values.append((f"{category_label} > {extension_label}", str(extension)))
    return values


affk.tag_subject_of(
    ("SEDABinaryDataObject", "file_category", "*"),
    {"choices": file_category_vocabulary},
)
rec.tag_subject_of(
    ("SEDABinaryDataObject", "file_category", "*"), {"novalue_label": " "}
)


class BinaryDataObjectFileInfoTab(viewlib.PrimaryTabWithoutBoxes):
    """Display file information of a binary data object"""

    __regid__ = "seda_bdo_file_information"
    __select__ = is_instance("SEDABinaryDataObject") & full_seda2_profile()

    rsection, display_ctrl = rtags_from_xsd_element("SEDABinaryDataObject", "FileInfo")


class PhysicalDataObjectTabbedPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("SEDAPhysicalDataObject")
    tabs = [
        "main_tab",
        _("seda_pdo_dimensions"),
        _("seda_do_relations"),
    ]


class PhysicalDataObjectDimensionsTab(viewlib.PrimaryTabWithoutBoxes):
    """Display physical dimensions of a physical data object"""

    __regid__ = "seda_pdo_dimensions"
    __select__ = is_instance("SEDAPhysicalDataObject")

    rsection, display_ctrl = rtags_from_xsd_element(
        "SEDAPhysicalDataObject", "PhysicalDimensions"
    )


class DataObjectRelationsTab(viewlib.PrimaryTabWithoutBoxes):
    """Display relations of a binary or physical data object"""

    __regid__ = "seda_do_relations"
    __select__ = (
        is_instance("SEDABinaryDataObject", "SEDAPhysicalDataObject")
        & full_seda2_profile()
    )

    _(
        "creating SEDARelationship (SEDARelationship seda_relationship "
        "SEDABinaryDataObject %(linkto)s)"
    )
    _(
        "creating SEDARelationship (SEDARelationship seda_relationship "
        "SEDAPhysicalDataObject %(linkto)s)"
    )

    def entity_call(self, entity):
        rschema = self._cw.vreg.schema.rschema("seda_relationship")
        if rschema.has_perm(self._cw, "add", toeid=entity.eid):
            urlparams = {"__redirectparams": "tab=" + self.__regid__}
            self.w(
                add_subobject_link(
                    entity,
                    "seda_relationship",
                    "object",
                    urlparams,
                    msg=self._cw._("add a SEDARelationship"),
                    klass="btn btn-success pull-right",
                )
            )
            self.w(tags.div(klass="clearfix"))
        rset = entity.related("seda_relationship", "object")
        if rset:
            self._cw.view(
                "list",
                rset=rset,
                parent=entity,
                w=self.w,
                tabid=self.__regid__,
                subvid="seda.listitem",
            )
        rset = entity.related("seda_target", "object")
        if rset:
            self.w(f"<h2>{self._cw._('Relationship target of')}</h2>")
            self.w(
                "<div>{}</div>".format(
                    self._cw._(
                        "This object is used as a relationship target of the following "
                        "entities"
                    )
                )
            )
            self._cw.view(
                "list", rset=rset, subvid="seda.relationship.reverse", w=self.w
            )


class RelationshipBusinessValueLinkEntityView(viewlib.BusinessValueLinkEntityView):
    __select__ = is_instance("SEDARelationship")

    def entity_value(self, entity):
        target = entity.seda_target[0] if entity.seda_target else None
        if target:
            value = tags.a(target.dc_title(), href=target.absolute_url())
        else:
            value = xml_escape(self._cw._("<no data-object specified>"))
        if entity.seda_type_relationship:
            concept = entity.seda_type_relationship[0]
            msg = self._cw._(", of relationship type %s") % concept.label()
        else:
            msg = self._cw._(", no relationship type specified")
        value += xml_escape(msg)
        return value


class CompressedBusinessValueEntityView(viewlib.BusinessValueEntityView):
    __select__ = is_instance("SEDACompressed")

    def entity_value(self, entity):
        if entity.compressed is None:
            value = self._cw.__("indifferent")
        else:
            value = self._cw.__("yes" if entity.compressed else "no")
        if entity.seda_algorithm:
            algorithm = entity.seda_algorithm[0].label()
            value += self._cw._(", using {algorithm}").format(algorithm=algorithm)
        return value


class RelationshipReverseEntityView(view.EntityView):
    __regid__ = "seda.relationship.reverse"
    __select__ = is_instance("SEDARelationship")

    def entity_call(self, entity):
        target = entity.seda_relationship[0]
        self.w(tags.a(target.dc_title(), href=target.absolute_url()))
        if entity.seda_type_relationship:
            concept = entity.seda_type_relationship[0]
            msg = self._cw._(", of relationship type %s") % concept.label()
        else:
            msg = self._cw._(", no relationship type specified")
        self.w(xml_escape(msg))


class UnitBusinessValueEntityView(viewlib.BusinessValueEntityView):
    __select__ = is_instance(
        "SEDAWidth",
        "SEDAHeight",
        "SEDADepth",
        "SEDADiameter",
        "SEDALength",
        "SEDAThickness",
        "SEDAWeight",
    )

    def entity_call(self, entity):
        super().entity_call(entity)
        if entity.seda_unit:
            unit = self._cw._("unit: {0}").format(entity.seda_unit[0].label())
        else:
            unit = self._cw._("<no unit specified>")
        self.w(f" ({xml_escape(unit)})")


class IBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Override adapter from compound when BDO is within a simplified profile to display the archive
    unit as parent.
    """

    __select__ = is_instance("SEDABinaryDataObject") & simplified_profile()

    def parent_entity(self):
        return self.entity.cw_adapt_to("ITreeBase").parent()



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda schema"""

from yams.buildobjs import RelationType, RelationDefinition
from yams.buildobjs import String

from cubicweb import _
from cubicweb.schema import ERQLExpression, RRQLExpression, RQLConstraint

from cubicweb_skos import schema as skos

from .xsd2yams import MULTIPLE_CHILDREN

_("1")
_("0..1")
_("0..n")
_("1..n")


def seda_profile_element(cardinalities=None, default_cardinality=None, annotable=True):
    """Class decorator adding attributes to configure a SEDA field."""

    def decorator(cls):
        if cardinalities and len(cardinalities) > 1:
            cls.add_relation(
                String(
                    required=True,
                    vocabulary=cardinalities,
                    default=default_cardinality,
                    internationalizable=True,
                ),
                name="user_cardinality",
            )
        if annotable:
            required = cls.__name__ in (
                "SEDAArchiveUnit",
                "SEDABinaryDataObject",
                "SEDAPhysicalDataObject",
            )
            if required:
                description = _(
                    "the first line will be used to display this "
                    "entity within the user interface"
                )
            else:
                description = None
            cls.add_relation(
                String(
                    fulltextindexed=True, required=required, description=description
                ),
                name="user_annotation",
            )
        return cls

    return decorator


class scheme_relation_type(RelationDefinition):
    """Special relation from a concept scheme to a relation type, that may be used to restrict
    possible concept of a particular relation without depending on the scheme's name or other kind
    of ambiguous reference mecanism.
    """

    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers",),
        "delete": ("managers",),
    }
    subject = "ConceptScheme"
    object = "CWRType"
    cardinality = "**"


class scheme_entity_type(RelationDefinition):
    """Special relation from a concept scheme to an entity type, when one add to constraint on a
    particular subject of a relation type (see `scheme_relation_type`).
    """

    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers",),
        "delete": ("managers",),
    }
    subject = "ConceptScheme"
    object = "CWEType"
    cardinality = "*?"


class seda_keyword_reference_to_scheme(RelationDefinition):
    subject = "SEDAKeywordReference"
    object = "ConceptScheme"
    cardinality = "?*"
    inlined = True


class file_category(RelationDefinition):
    subject = "SEDABinaryDataObject"
    object = "Concept"
    cardinality = "**"
    constraints = [
        RQLConstraint(
            'O in_scheme CS, CS scheme_relation_type CR, CR name "file_category"'
        ),
    ]


class ordering(RelationDefinition):
    subject = tuple(etype for etype, _ in MULTIPLE_CHILDREN)
    object = "Int"
    cardinality = "?1"


class container(RelationType):
    inlined = False


class _clone_of(RelationDefinition):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": (
            "managers",
            "users",
        ),
        "delete": (),
    }
    name = "clone_of"
    cardinality = "?*"
    inlined = True


class clone_of_archive_transfer(_clone_of):
    subject = "SEDAArchiveTransfer"
    object = "SEDAArchiveTransfer"


class clone_of_archive_unit(_clone_of):
    subject = "SEDAArchiveUnit"
    object = "SEDAArchiveUnit"


class title(RelationDefinition):
    subject = "SEDAArchiveTransfer"
    object = "String"
    description = _("title for this profile, not used in the generated schema")
    cardinality = "11"
    fulltextindexed = True


class simplified_profile(RelationDefinition):
    subject = "SEDAArchiveTransfer"
    object = "Boolean"
    default = False
    description = _(
        "simplified profiles are compatible with older version of SEDA, but have not "
        "the full SEDA 2 expressivness"
    )
    cardinality = "11"


class wrap_dataobjects(RelationDefinition):
    subject = "SEDAArchiveTransfer"
    object = "Boolean"
    default = False
    description = _(
        "check the option if you want BinaryDataObject and PhysicalDataObject "
        "to be wrapped in a DataObjectGroup object"
    )
    cardinality = "11"


class code_keyword_type(RelationDefinition):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", RRQLExpression("U has_update_permission S")),
        "delete": ("managers", RRQLExpression("U has_update_permission S")),
    }
    subject = "ConceptScheme"
    object = "Concept"
    cardinality = "?*"
    inlined = True
    description = _("SEDA code keyword type for this scheme")
    constraints = [
        RQLConstraint(
            "O in_scheme CS, CS scheme_relation_type RT, "
            'RT name "seda_keyword_type_to"'
        )
    ]


class compat_list(RelationDefinition):
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (),
        "update": (),
    }
    subject = "SEDAArchiveTransfer"
    object = "String"
    description = _(
        "names of format in which the profile may be exported (e.g. 'SEDA 2, SEDA 1')"
    )


language_code = skos.Label.get_relation("language_code")
language_code.constraints[0].max = 7


def post_build_callback(schema):
    from cubicweb_seda import seda_profile_container_def, iter_all_rdefs

    container_etypes = ("SEDAArchiveTransfer", "SEDAArchiveUnit")
    for etype, parent_rdefs in seda_profile_container_def(schema):
        # add relation to the container from every entity type within the compound graph
        for container_etype in container_etypes:
            # special case for SEDAArchiveUnit's cardinality as it may be both
            # contained and container
            container_rdef = RelationDefinition(
                etype,
                "container",
                container_etype,
                cardinality="?*" if etype == "SEDAArchiveUnit" else "1*",
                __permissions__={
                    "add": (),
                    "delete": (),
                    "read": ("managers", "users", "guests"),
                },
            )
            schema.add_relation_def(container_rdef)
        eschema = schema[etype]
        # set permissions on entity types from the compound graph according to permission on the
        # container entity
        for action in ("update", "delete"):
            eschema.set_action_permissions(
                action,
                (ERQLExpression(f"U has_{action}_permission C, X container C"),),
            )
    for action in ("update", "delete"):
        schema["SEDAArchiveUnit"].set_action_permissions(
            action,
            (
                ERQLExpression(f"U has_{action}_permission C, X container C"),
                ERQLExpression(
                    "NOT EXISTS(X container C), U in_group G, "
                    'G name IN ("managers", "users")'
                ),
            ),
        )
    # set permissions on all relation defs related to the compound graph according to permission on
    # the container entity
    for rdef, role in iter_all_rdefs(schema, "SEDAArchiveTransfer"):
        if role == "subject":
            target_etype, var = rdef.subject, "S"
        else:
            target_etype, var = rdef.object, "O"
        rrql_exprs = []
        if target_etype == "SEDAArchiveUnit":
            rrql_exprs.append(
                "U has_update_permission {0}, NOT EXISTS({0} container C)".format(var)
            )
            rrql_exprs.append(f"U has_update_permission C, {var} container C")

        else:
            if target_etype == "SEDAArchiveTransfer":
                rrql_exprs.append(f"U has_update_permission {var}")
            else:
                rrql_exprs.append(f"U has_update_permission C, {var} container C")
        permissions = [RRQLExpression(expr) for expr in rrql_exprs]
        for action in ("add", "delete"):
            rdef.set_action_permissions(action, permissions)



# copyright 2015-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda views related to XSD download"""

from logilab.common.registry import objectify_predicate

from cubicweb import view, _
from cubicweb.predicates import is_instance, one_line_rset
from cubicweb_web import httpcache, action
from cubicweb_web.views import idownloadable

DEFAULT_VERSION = "2.1"


@objectify_predicate
def format_supported(cls, req, rset=None, entity=None, **kwargs):
    """Predicate matching cases where an expected format (specified as `version` form param or as a
    `seda_version` class attribute on the appobject) is supported by the context's transfer entity.
    """
    if entity is None:
        entity = rset.get_entity(0, 0)
    version = getattr(
        cls,
        "seda_version",
        kwargs.get("version", req.form.get("version", DEFAULT_VERSION)),
    )
    if version != DEFAULT_VERSION and not entity.simplified_profile:
        return 0
    return int(("SEDA " + version) in entity.formats_compat)


class SEDADownloadAction(action.Action):
    __abstract__ = True
    __select__ = (
        action.Action.__select__
        & one_line_rset()
        & is_instance("SEDAArchiveTransfer")
        & format_supported()
    )
    category = "moreactions"

    # set those in concret classes
    __regid__ = title = seda_version = export_format = order = None

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(
            vid="seda.export", version=self.seda_version, format=self.export_format
        )


class SEDA20DownloadRNGAction(SEDADownloadAction):
    title = _("SEDA 2.1 RNG export")
    order = 100
    seda_version = "2.1"
    export_format = "rng"
    __regid__ = f"seda.export.{seda_version}.{export_format}"


class SEDA1DownloadRNGAction(SEDADownloadAction):
    title = _("SEDA 1.0 RNG export")
    order = 102
    seda_version = "1.0"
    export_format = "rng"
    __regid__ = f"seda.export.{seda_version}.{export_format}"


class SEDA02DownloadRNGAction(SEDADownloadAction):
    title = _("SEDA 0.2 RNG export")
    order = 104
    seda_version = "0.2"
    export_format = "rng"
    __regid__ = f"seda.export.{seda_version}.{export_format}"


class SEDA2DownloadHTMLAction(SEDADownloadAction):
    title = _("HTML documentation")
    order = 110
    seda_version = "2.1"
    export_format = "html"
    __regid__ = f"seda.export.{seda_version}.{export_format}"


class SEDAExportViewMixin:
    __regid__ = "seda.export"
    __select__ = one_line_rset() & is_instance("SEDAArchiveTransfer")


class SEDAExportView(SEDAExportViewMixin, idownloadable.DownloadView):
    """SEDA archive transfer export view, to download rng, html or xsd
    representation of a profile."""

    __select__ = SEDAExportViewMixin.__select__ & format_supported()

    http_cache_manager = httpcache.NoHTTPCacheManager

    @property
    def seda_adapter_id(self):
        return "SEDA-{}.{}".format(
            self.cw_extra_kwargs.get(
                "version", self._cw.form.get("version", DEFAULT_VERSION)
            ),
            self.cw_extra_kwargs.get(
                "format", self._cw.form.get("format", "rng")
            ).lower(),
        )

    def set_request_content_type(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to(self.seda_adapter_id)
        filename = "{}-{}.{}".format(
            entity.dc_title(),
            self._cw.form.get("version", DEFAULT_VERSION),
            self._cw.form.get("format", "rng").lower(),
        )
        self._cw.set_content_type(
            adapter.content_type,
            filename=filename,
            encoding=adapter.encoding,
            disposition="attachment",
        )

    def call(self, **kwargs):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to(self.seda_adapter_id)
        self.w(adapter.dump())


class SEDANonSupportedExportView(SEDAExportViewMixin, view.EntityView):
    """SEDA archive transfer export view for cases where an unsupported export is specified."""

    __select__ = SEDAExportViewMixin.__select__ & ~format_supported()

    def entity_call(self, entity):
        self.w('<p class="bg-error">')
        version = self.cw_extra_kwargs.get(
            "version", self._cw.form.get("version", DEFAULT_VERSION)
        )
        self.w(
            self._cw._("Export to SEDA version {0} is not possible.").format(version)
        )
        self.w("</p>")
        self.w("<p>")
        self.w(self._cw._("Supported formats: {0}.").format(entity.compat_list))
        self.w("</p>")



#
# Copyright © LOGILAB S.A. (Paris, FRANCE) 2016-2021
# Contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systemsand/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-C license and that you accept its terms.
#
"""THIS FILE IS NOT GENERATED FROM SEDA XSD FILES, IT CAN BE EDITED"""

from cubicweb.web.views import uicfg

afs = uicfg.autoform_section


# form

afs.tag_object_of(("*", "seda_data_object_group", "*"), "main", "hidden")
afs.tag_object_of(("*", "seda_log_book", "SEDAArchiveTransfer"), "main", "hidden")
afs.tag_object_of(("*", "seda_data_object_group_system_id", "*"), "main", "hidden")
afs.tag_subject_of(("*", "file_category", "*"), "main", "hidden")



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Miscellaneous test utilities."""

from contextlib import contextmanager

from cubicweb import ValidationError, Unauthorized
from cubicweb.rset import NoResultError


@contextmanager
def assertValidationError(self, cnx):
    with self.assertRaises(ValidationError) as cm:
        yield cm
        cnx.commit()
    cnx.rollback()


@contextmanager
def assertUnauthorized(self, cnx):
    with self.assertRaises(Unauthorized) as cm:
        yield cm
        cnx.commit()
    cnx.rollback()


def create_transfer_to_bdo(cnx):
    """Create minimal ArchiveTransfer down to a BinaryDataObject and return the later."""
    transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
    bdo = create_data_object(transfer)
    # commit and clear cache to allow access to container relation
    cnx.commit()
    bdo.cw_clear_all_caches()
    return bdo


def create_archive_unit(parent, archive_unit_reference=False, title=None, **kwargs):
    """Create an archive unit and its mandatory children:

    return (archive unit, alternative, sequence) if archive_unit_reference is false (the default),
    else (archive unit, alternative, reference).
    """
    cnx = kwargs.pop("cnx", getattr(parent, "_cw", None))
    kwargs.setdefault("user_annotation", "archive unit title")
    au = cnx.create_entity("SEDAArchiveUnit", seda_archive_unit=parent, **kwargs)
    alt = cnx.create_entity(
        "SEDAAltArchiveUnitArchiveUnitRefId",
        reverse_seda_alt_archive_unit_archive_unit_ref_id=au,
    )
    if archive_unit_reference:
        last = cnx.create_entity(
            "SEDAArchiveUnitRefId", seda_archive_unit_ref_id_from=alt
        )
    else:
        try:
            concept = cnx.execute('Concept C WHERE L label_of C, L label "file"').one()
        except NoResultError:
            with cnx.security_enabled(write=False):
                scheme = scheme_for_type(cnx, "seda_description_level_to", None, "file")
                # Concept
                concept = scheme.reverse_in_scheme[0]
        last = cnx.create_entity(
            "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement",
            reverse_seda_seq_alt_archive_unit_archive_unit_ref_id_management=alt,
        )
        cnx.create_entity(
            "SEDADescriptionLevel",
            seda_description_level_from=last,
            seda_description_level_to=concept,
        )
        cnx.create_entity("SEDATitle", seda_title=last, title=title)

    return au, alt, last


def create_data_object(parent, **kwargs):
    cnx = getattr(parent, "_cw", None)
    etype = "SEDABinaryDataObject"
    relation_type = "seda_binary_data_object"

    if "seda_physical_data_object" in kwargs:
        etype = "SEDAPhysicalDataObject"
        relation_type = "seda_physical_data_object"

    if parent.cw_etype == "SEDAArchiveTransfer":
        kwargs[relation_type] = parent
    kwargs.setdefault("user_annotation", "data object title")
    bdo = cnx.create_entity(etype, **kwargs)
    if etype == "SEDABinaryDataObject":
        choice = cnx.create_entity(
            "SEDAAltBinaryDataObjectAttachment",
            reverse_seda_alt_binary_data_object_attachment=bdo,
        )
        cnx.create_entity(
            "SEDAAttachment", seda_attachment=choice
        )  # Choice cannot be empty
    if parent.cw_etype != "SEDAArchiveTransfer":
        cnx.create_entity(
            "SEDADataObjectReference",
            user_cardinality="0..n",
            seda_data_object_reference=parent,
            seda_data_object_reference_id=bdo,
        )

    return bdo


def create_authority_record(cnx, name, kind="person", **kwargs):
    """Return an AuthorityRecord with specified kind and name."""
    kind_eid = cnx.find("AgentKind", name=kind)[0][0]
    record = cnx.create_entity("AuthorityRecord", agent_kind=kind_eid, **kwargs)
    cnx.create_entity(
        "NameEntry", parts=name, form_variant="authorized", name_entry_for=record
    )
    return record


def map_cs_to_type(scheme, rtype, etype=None):
    cnx = scheme._cw
    cnx.execute(
        "SET CS scheme_relation_type RT WHERE CS eid %(cs)s, RT name %(rt)s",
        {"cs": scheme.eid, "rt": rtype},
    )
    if etype is not None:
        cnx.execute(
            "SET CS scheme_entity_type ET WHERE CS eid %(cs)s, ET name %(et)s",
            {"cs": scheme.eid, "et": etype},
        )


# extra kwargs useful for client lib (e.g. saem)
def scheme_for_type(cnx, rtype, etype, *concept_labels, **kwargs):
    """Create a concept scheme an map it to give rtype and optional etype."""
    scheme = cnx.create_entity(
        "ConceptScheme", title=f"{rtype}/{etype} vocabulary", **kwargs
    )
    map_cs_to_type(scheme, rtype, etype)
    for label in concept_labels:
        scheme.add_concept(label)
    return scheme


def scheme_for_rtype(cnx, rtype, *concept_labels):
    return scheme_for_type(cnx, rtype, None, *concept_labels)
